# -*- coding: utf-8 -*-
"""Test suite for axonius_api_client."""
import tempfile

import pytest

from axonius_api_client.exceptions import ToolsError
from axonius_api_client.tools import (
    coerce_bool,
    coerce_int,
    datetime,
    dt_min_ago,
    dt_now,
    dt_parse,
    dt_within_min,
    get_path,
    grouper,
    is_int,
    join_url,
    json_dump,
    json_load,
    json_reload,
    listify,
    path_read,
    path_write,
    pathlib,
    strip_left,
    strip_right,
    timedelta,
)

from ..utils import IS_WINDOWS


'''
class TestValType:
    """Test axonius_api_client.val_type."""

    def test_good(self):
        """Pass."""
        val_type(1, INT)

    def test_bad(self):
        """Pass."""
        with pytest.raises(ToolsError):
            val_type("", INT)
'''


class TestCoerce:
    """Test axonius_api_client.join_url."""

    def test_int(self):
        """Pass."""
        with pytest.raises(ToolsError):
            coerce_int("badwolf")

        assert coerce_int("456") == 456

    def test_bool(self):
        """Pass."""
        with pytest.raises(ToolsError):
            coerce_bool("badwolf")

        assert coerce_bool("y") is True
        assert coerce_bool("yes") is True
        assert coerce_bool("true") is True
        assert coerce_bool("True") is True
        assert coerce_bool("1") is True
        assert coerce_bool(1) is True
        assert coerce_bool("t") is True
        assert coerce_bool(True) is True
        assert coerce_bool("n") is False
        assert coerce_bool("no") is False
        assert coerce_bool("false") is False
        assert coerce_bool("False") is False
        assert coerce_bool("0") is False
        assert coerce_bool(0) is False
        assert coerce_bool("f") is False
        assert coerce_bool(False) is False


class TestJoinUrl:
    """Test axonius_api_client.join_url."""

    def test_url(self):
        """Test url gets joined properly no matter the slashes."""
        r = join_url("https://test.com")
        assert r == "https://test.com/"
        r = join_url("https://test.com/")
        assert r == "https://test.com/"
        r = join_url("https://test.com////")
        assert r == "https://test.com/"
        r = join_url("https://test.com", "")
        assert r == "https://test.com/"
        r = join_url("https://test.com", "", "")
        assert r == "https://test.com/"
        r = join_url("https://test.com", "/", "")
        assert r == "https://test.com/"
        r = join_url("https://test.com", "/", "/")
        assert r == "https://test.com/"

    def test_url_path(self):
        """Test url, path gets joined properly no matter the slashes."""
        r = join_url("https://test.com", "a")
        assert r == "https://test.com/a"
        r = join_url("https://test.com", "/a")
        assert r == "https://test.com/a"
        r = join_url("https://test.com", "//a")
        assert r == "https://test.com/a"
        r = join_url("https://test.com", "a/")
        assert r == "https://test.com/a/"
        r = join_url("https://test.com", "a/b")
        assert r == "https://test.com/a/b"
        r = join_url("https://test.com", "a/b", "")
        assert r == "https://test.com/a/b"
        r = join_url("https://test.com", "a/b/", "")
        assert r == "https://test.com/a/b/"
        r = join_url("https://test.com", "a/b", "/")
        assert r == "https://test.com/a/b/"
        r = join_url("https://test.com", "a/b", "/////")
        assert r == "https://test.com/a/b/"

    def test_url_path_route(self):
        """Test url, path, route gets joined properly no matter the slashes."""
        r = join_url("https://test.com", "a", "b")
        assert r == "https://test.com/a/b"
        r = join_url("https://test.com", "/a", "b")
        assert r == "https://test.com/a/b"
        r = join_url("https://test.com", "//a", "b")
        assert r == "https://test.com/a/b"
        r = join_url("https://test.com", "a", "b/c/d")
        assert r == "https://test.com/a/b/c/d"


'''
class TestJoinDot:
    """Test axonius_api_client.join_dot."""

    def test_multi(self):
        """Test dot join multi."""
        r = join_dot(obj=["x", "a", "c"])
        assert r == "x.a.c"

    def test_multi_with_empty_false(self):
        """Test dot join multi with empty=False."""
        r = join_dot(obj=["x", "", "a", None, "c", []], empty=False)
        assert r == "x.a.c"

    def test_multi_with_empty_true(self):
        """Test dot join multi with empty=True."""
        r = join_dot(obj=["x", "", "a", None, "c"], empty=True)
        assert r == "x..a.None.c"

    def test_single(self):
        """Test dot join single."""
        r = join_dot(obj=["x"])
        assert r == "x"

    def test_non_list(self):
        """Test dot join single."""
        r = join_dot(obj="x")
        assert r == "x"

    def test_empty_list(self):
        """Test dot join single."""
        r = join_dot(obj=[])
        assert r == ""

    def test_single_none(self):
        """Test dot join with None."""
        r = join_dot(obj=None)
        assert r == ""


class TestJoinComma:
    """Test axonius_api_client.join_dot."""

    def test_multi(self):
        """Test comma join multi."""
        r = join_comma(obj=["x", "a", "c"])
        assert r == "x, a, c"

    def test_multi_no_indent(self):
        """Test comma join multi with indent=False."""
        r = join_comma(obj=["x", "a", "c"], indent=False)
        assert r == "x,a,c"

    def test_multi_with_empty_false(self):
        """Test comma join multi with empty=False."""
        r = join_comma(obj=["x", "", "a", None, "c", []], empty=False)
        assert r == "x, a, c"

    def test_multi_with_empty_true(self):
        """Test comma join list with multi items with empty=True."""
        r = join_comma(obj=["x", "", "a", None, "c"], empty=True)
        assert r == "x, , a, None, c"

    def test_single(self):
        """Test comma join list with single item."""
        r = join_comma(obj=["x"])
        assert r == "x"

    def test_non_list(self):
        """Test comma non list."""
        r = join_comma(obj="x")
        assert r == "x"

    def test_empty_list(self):
        """Test comma empty list."""
        r = join_comma(obj=[])
        assert r == ""

    def test_single_none(self):
        """Test comma join with None."""
        r = join_comma(obj=None)
        assert r == ""


class TestJoinCr:
    """Test axonius_api_client.join_cr."""

    def test_multi(self):
        """Test cr join multi."""
        r = join_cr(obj=["x", "a", "c"])
        assert r == "\n  x\n  a\n  c"

    def test_single(self):
        """Test cr join multi."""
        r = join_cr(obj=["x"])
        assert r == "\n  x"

    def test_single_non_list(self):
        """Test cr join list w/ single entry."""
        r = join_cr(obj="x")
        assert r == "\n  x"

    def test_single_empty_list(self):
        """Test cr join empty list."""
        r = join_cr(obj=[])
        assert r == ""

    def test_single_none(self):
        """Test cr join with None."""
        r = join_cr(obj=None)
        assert r == ""

    def test_post_and_pre(self):
        """Test cr join with post and pre = True."""
        r = join_cr(obj=["x", "a", "c"], pre=True, post=True)
        assert r == "\n  x\n  a\n  c\n  "
'''


class TestGetPath:
    """Test axonius_api_client.get_path."""

    def test_str(self):
        """Test resolve with a string."""
        r = get_path(obj="/../badwolf")
        assert isinstance(r, pathlib.Path)
        if IS_WINDOWS:
            assert format(r).endswith("badwolf")
        else:
            assert format(r) == format("/badwolf")

    def test_pathlib(self):
        """Test resolve with a pathlib.Path."""
        r = get_path(obj=pathlib.Path("."))
        assert isinstance(r, pathlib.Path)

    def test_user(self):
        """Test resolve with ~."""
        r = get_path(obj="~")
        assert isinstance(r, pathlib.Path)
        assert format(r) == format(pathlib.Path.home())


class TestPathWrite:
    """Test axonius_api_client.path_write."""

    def test_simple_pathlib(self, tmp_path):
        """Test simple write with pathlib object."""
        sub1 = tmp_path / "sub1"
        sub2 = sub1 / "sub2"
        path = sub2 / "file.txt"
        data = "abc\n123\n"
        ret_path, ret_write = path_write(obj=path, data=data)
        assert ret_path.read_text() == data
        assert format(ret_path) == format(path)
        assert ret_write == len(data)
        if IS_WINDOWS:
            assert ret_path.stat().st_mode == 33206
            assert ret_path.parent.stat().st_mode == 16895
        else:
            assert ret_path.stat().st_mode == 33152
            assert ret_path.parent.stat().st_mode == 16832

    def test_simple_str(self, tmp_path):
        """Test simple write with path as str."""
        sub1 = tmp_path / "sub1"
        sub2 = sub1 / "sub2"
        path = sub2 / "file.txt"
        data = "abc\n123\n"
        ret_path, ret_write = path_write(obj=format(path), data=data)
        assert ret_path.read_text() == data
        assert format(ret_path) == format(path)
        assert ret_write == len(data)

    def test_parent_fail(self, tmp_path):
        """Test simple write with pathlib object."""
        sub1 = tmp_path / "sub1"
        sub2 = sub1 / "sub2"
        path = sub2 / "file.txt"
        data = "abc\n123\n"
        with pytest.raises(ToolsError):
            path_write(obj=path, data=data, make_parent=False)

    def test_noperm_parent(self):
        """Pass."""
        tmpdir = pathlib.Path(tempfile.gettempdir())
        path = tmpdir / "file.txt"
        data = "abc\n123\n"
        ret_path, ret_write = path_write(obj=path, data=data, overwrite=True)
        assert ret_path.read_text() == data

    def test_overwrite_false(self, tmp_path):
        """Test overwrite=False."""
        sub1 = tmp_path / "sub1"
        sub2 = sub1 / "sub2"
        path = sub2 / "file.txt"
        data = "abc\n123\n"
        path_write(obj=path, data=data)
        with pytest.raises(ToolsError):
            path_write(obj=path, data=data, overwrite=False)

    def test_overwrite_true(self, tmp_path):
        """Test overwrite=True."""
        sub1 = tmp_path / "sub1"
        sub2 = sub1 / "sub2"
        path = sub2 / "file.txt"
        data = "abc\n123\n"
        path_write(obj=path, data=data)
        path_write(obj=path, data=data, overwrite=True)
        assert path.is_file()

    def test_binary_true_nonbinary(self, tmp_path):
        """Test binary=True with nonbinary data."""
        sub1 = tmp_path / "sub1"
        sub2 = sub1 / "sub2"
        path = sub2 / "file.txt"
        data = "abc\n123\n"
        ret_path, ret_write = path_write(obj=path, data=data, binary=True)
        assert ret_path.read_text() == "abc\n123\n"
        if IS_WINDOWS:
            assert ret_path.read_bytes() == b"abc\n123\n"
        else:
            assert ret_path.read_bytes() == b"abc\n123\n"

    def test_binary_true_binary(self, tmp_path):
        """Test binary=True with binary data."""
        sub1 = tmp_path / "sub1"
        sub2 = sub1 / "sub2"
        path = sub2 / "file.txt"
        data = b"abc\n123\n"
        ret_path, ret_write = path_write(obj=path, data=data, binary=True)
        assert ret_path.read_text() == "abc\n123\n"
        if IS_WINDOWS:
            assert ret_path.read_bytes() == b"abc\n123\n"
        else:
            assert ret_path.read_bytes() == b"abc\n123\n"

    def test_binary_false_nonbinary(self, tmp_path):
        """Test binary=False with nonbinary data."""
        sub1 = tmp_path / "sub1"
        sub2 = sub1 / "sub2"
        path = sub2 / "file.txt"
        data = "abc\n123\n"
        ret_path, ret_write = path_write(obj=path, data=data, binary=False)
        assert ret_path.read_text() == "abc\n123\n"
        if IS_WINDOWS:
            assert ret_path.read_bytes() == b"abc\r\n123\r\n"
        else:
            assert ret_path.read_bytes() == b"abc\n123\n"

    def test_binary_false_binary(self, tmp_path):
        """Test binary=False with binary data."""
        sub1 = tmp_path / "sub1"
        sub2 = sub1 / "sub2"
        path = sub2 / "file.txt"
        data = b"abc\n123\n"
        ret_path, ret_write = path_write(obj=path, data=data, binary=False)
        assert ret_path.read_text() == "abc\n123\n"
        if IS_WINDOWS:
            assert ret_path.read_bytes() == b"abc\r\n123\r\n"
        else:
            assert ret_path.read_bytes() == b"abc\n123\n"

    def test_is_json_false_dotjson_nonjson(self, tmp_path):
        """Test is_json=False with .json in filename and invalid json data."""
        sub1 = tmp_path / "sub1"
        sub2 = sub1 / "sub2"
        path = sub2 / "file.json"
        data = b""
        ret_path, ret_write = path_write(obj=path, data=data, is_json=False)
        assert ret_path.read_text() == data.decode()

    def test_is_json_false_dotjson_json(self, tmp_path):
        """Test is_json=False with .json in filename and valid json data."""
        sub1 = tmp_path / "sub1"
        sub2 = sub1 / "sub2"
        path = sub2 / "file.json"
        data = {"x": 2}
        ret_path, ret_write = path_write(obj=path, data=data, is_json=False)
        assert ret_path.read_text() == '{\n  "x": 2\n}'

    def test_is_json_true_nonjson(self, tmp_path):
        """Test is_json=True with .json not in filename and invalid json data."""
        sub1 = tmp_path / "sub1"
        sub2 = sub1 / "sub2"
        path = sub2 / "file.text"
        data = pytest
        with pytest.raises(Exception):
            path_write(obj=path, data=data, is_json=True)

    def test_is_json_true_json(self, tmp_path):
        """Test is_json=True with .json not in filename and valid json data."""
        sub1 = tmp_path / "sub1"
        sub2 = sub1 / "sub2"
        path = sub2 / "file.text"
        data = {"x": 2}
        ret_path, ret_write = path_write(obj=path, data=data, is_json=True)
        assert ret_path.read_text() == '{\n  "x": 2\n}'


class TestPathRead:
    """Test axonius_api_client.path_read."""

    def test_pathlib(self, tmp_path):
        """Test simple write with pathlib object."""
        sub1 = tmp_path / "sub1"
        sub2 = sub1 / "sub2"
        path = sub2 / "file.txt"
        data = "abc\n123\n"
        wret_path, ret_write = path_write(obj=path, data=data)
        rret_path, ret_read = path_read(obj=path)
        assert wret_path == rret_path
        assert ret_read == data

    def test_str(self, tmp_path):
        """Test simple write with str."""
        sub1 = tmp_path / "sub1"
        sub2 = sub1 / "sub2"
        path = sub2 / "file.txt"
        data = "abc\n123\n"
        wret_path, ret_write = path_write(obj=path, data=data)
        rret_path, ret_read = path_read(obj=format(path))
        assert wret_path == rret_path
        assert ret_read == data

    def test_binary_true(self, tmp_path):
        """Test binary=True."""
        sub1 = tmp_path / "sub1"
        sub2 = sub1 / "sub2"
        path = sub2 / "file.txt"
        data = "abc\n123\n"
        wret_path, ret_write = path_write(obj=path, data=data)
        rret_path, ret_read = path_read(obj=format(path), binary=True)
        assert wret_path == rret_path
        if IS_WINDOWS:
            assert ret_read == b"abc\r\n123\r\n"
        else:
            assert ret_read == b"abc\n123\n"

    def test_binary_false(self, tmp_path):
        """Test binary=False."""
        sub1 = tmp_path / "sub1"
        sub2 = sub1 / "sub2"
        path = sub2 / "file.txt"
        data = "abc\n123\n"
        wret_path, ret_write = path_write(obj=path, data=data)
        rret_path, ret_read = path_read(obj=format(path), binary=False)
        assert wret_path == rret_path
        assert ret_read == data

    def test_is_json_true_nonjson(self, tmp_path):
        """Test is_json=True with .json not in filename and invalid json data."""
        sub1 = tmp_path / "sub1"
        sub2 = sub1 / "sub2"
        path = sub2 / "file.text"
        data = "abc\n123\n"
        wret_path, ret_write = path_write(obj=path, data=data)
        with pytest.raises(Exception):
            path_read(obj=path, is_json=True)

    def test_is_json_false_dotjson_nonjson(self, tmp_path):
        """Test is_json=False with .json in filename and invalid json data."""
        sub1 = tmp_path / "sub1"
        sub2 = sub1 / "sub2"
        path = sub2 / "file.json"
        data = "abc\n123\n"
        wret_path, ret_write = path_write(obj=path, data=data)
        rret_path, ret_read = path_read(obj=path, is_json=False)
        assert wret_path == rret_path
        assert ret_read == data

    def test_is_json_false_dotjson_json(self, tmp_path):
        """Test is_json=False with .json in filename and valid json data."""
        sub1 = tmp_path / "sub1"
        sub2 = sub1 / "sub2"
        path = sub2 / "file.json"
        data = {"x": 2}
        wret_path, ret_write = path_write(obj=path, data=data)
        rret_path, ret_read = path_read(obj=path, is_json=False)
        assert wret_path == rret_path
        assert ret_read == data

    def test_not_found(self, tmp_path):
        """Test is_json=False with .json in filename and valid json data."""
        sub1 = tmp_path / "sub1"
        sub2 = sub1 / "sub2"
        path = sub2 / "file.json"
        with pytest.raises(ToolsError):
            path_read(obj=path)


class TestGrouper:
    """Test grouper."""

    def test_iter(self):
        """Simple test."""
        x = list(grouper([1, 2, 3, 4, 5, 6], 2))
        assert x == [(1, 2), (3, 4), (5, 6)]

    def test_iter_off1(self):
        """Simple test."""
        x = list(grouper([1, 2, 3, 4, 5, 6, 7], 2))
        assert x == [(1, 2), (3, 4), (5, 6), (7, None)]

    def test_iter_off1_strfill(self):
        """Simple test."""
        x = list(grouper([1, 2, 3, 4, 5, 6, 7], 2, "x"))
        assert x == [(1, 2), (3, 4), (5, 6), (7, "x")]


'''
class TestNestDepth:
    """Test listify."""

    def test_dict1(self):
        """Simple test."""
        x = nest_depth(obj={"x": 1, "y": 1})
        assert x == 1

    def test_dict2(self):
        """Simple test."""
        x = nest_depth(obj={"x": 1, "y": {"z": 1}})
        assert x == 2

    def test_tuple(self):
        """Simple test."""
        x = nest_depth(obj=(1, 2))
        assert x == 1

    def test_list1(self):
        """Simple test."""
        x = nest_depth(obj=[1, 2])
        assert x == 1

    def test_list2(self):
        """Simple test."""
        x = nest_depth(obj=[[1], [2]])
        assert x == 2

    @pytest.mark.parametrize("val", [1, "1", None], scope="class")
    def test_not_complex(self, val):
        """Simple test."""
        x = nest_depth(obj=val)
        assert x == 0

    def test_mix(self):
        """Simple test."""
        x = nest_depth(obj=["1", ["a", {}]])
        assert x == 3

'''


class TestListify:
    """Test listify."""

    def test_dict_keys(self):
        """Simple test."""
        x = listify(obj={"x": 1, "y": 1}, dictkeys=True)
        assert x == ["x", "y"]

    def test_dict_notkeys(self):
        """Simple test."""
        x = listify(obj={"x": 1, "y": 1}, dictkeys=False)
        assert x == [{"x": 1, "y": 1}]

    def test_tuple(self):
        """Simple test."""
        x = listify(obj=(1, 2))
        assert x == [1, 2]

    def test_list(self):
        """Simple test."""
        x = listify(obj=[1, 2])
        assert x == [1, 2]

    def test_int(self):
        """Simple test."""
        x = listify(obj=1)
        assert x == [1]

    def test_str(self):
        """Simple test."""
        x = listify(obj="1")
        assert x == ["1"]

    def test_none(self):
        """Simple test."""
        x = listify(obj=None)
        assert x == []


class TestIsInt:
    """Test is_*."""

    @pytest.mark.parametrize("ok", [0, 4], scope="class")
    @pytest.mark.parametrize("bad", ["1", False, True, b"1"], scope="class")
    def test_int_digit_false(self, ok, bad):
        """Simple test."""
        assert is_int(obj=ok, digit=False)
        assert not is_int(obj=bad, digit=False)

    @pytest.mark.parametrize("ok", [0, 4, "1", b"1"], scope="class")
    @pytest.mark.parametrize("bad", [False, True, {}, "x", b"x"], scope="class")
    def test_int_digit_true(self, ok, bad):
        """Simple test."""
        assert is_int(obj=ok, digit=True)
        assert not is_int(obj=bad, digit=True)


class TestStripLeft:
    """Test strip_left."""

    def test_left_str(self):
        """Simple test."""
        x = "badwolfbadwolf"
        y = strip_left(obj=x, fix="badwolf")
        assert y == "badwolf"

    def test_left_list(self):
        """Simple test."""
        x = ["badwolfbadwolf", "badwolf", "abadwolfbadwolf"]
        y = strip_left(obj=x, fix="badwolf")
        assert y == ["badwolf", "", "abadwolfbadwolf"]

    def test_left_str_nomatch(self):
        """Simple test."""
        x = "abadwolfbadwolf"
        y = strip_left(obj=x, fix="badwolf")
        assert y == "abadwolfbadwolf"


class TestStripRight:
    """Test strip_right."""

    def test_right_str(self):
        """Simple test."""
        x = "badwolfbadwolf"
        y = strip_right(obj=x, fix="badwolf")
        assert y == "badwolf"

    def test_right_str_nomatch(self):
        """Simple test."""
        x = "badwolfbadwolfa"
        y = strip_right(obj=x, fix="badwolf")
        assert y == "badwolfbadwolfa"

    def test_right_list(self):
        """Simple test."""
        x = ["badwolfbadwolf", "badwolf", "badwolfbadwolfa"]
        y = strip_right(obj=x, fix="badwolf")
        assert y == ["badwolf", "", "badwolfbadwolfa"]


class TestJsonLoad:
    """Test json_load."""

    def test_load(self):
        """Simple test."""
        x = "{}"
        y = json_load(obj=x)
        assert y == {}

    def test_load_error_false(self):
        """Simple test."""
        x = "xxx"
        y = json_load(obj=x, error=False)
        assert y == "xxx"

    def test_load_error_true(self):
        """Simple test."""
        x = "xxx"
        with pytest.raises(Exception):
            json_load(obj=x, error=True)


class TestJsonDump:
    """Test json_dump."""

    def test_dump(self):
        """Simple test."""
        x = {"x": 2}
        y = json_dump(obj=x)
        assert y == '{\n  "x": 2\n}'

    def test_dump_error_false(self):
        """Simple test."""
        x = pytest
        y = json_dump(obj=x, error=False)
        assert y == pytest

    def test_dump_error_true(self):
        """Simple test."""
        x = pytest
        with pytest.raises(Exception):
            json_dump(obj=x, error=True)


class TestJsonReload:
    """Test json_dump."""

    def test_re_load(self):
        """Simple test."""
        x = '{"x": 2}'
        y = json_reload(obj=x)
        assert y == '{\n  "x": 2\n}'

    def test_re_load_error_false(self):
        """Simple test."""
        x = "{"
        y = json_reload(obj=x, error=False)
        assert y == x

    def test_re_load_error_true(self):
        """Simple test."""
        x = "{"
        with pytest.raises(Exception):
            json_reload(obj=x, error=True)


class TestDtMinAgo:
    """Test dt_*."""

    def test_min_ago_utc_str(self):
        """Simple test."""
        then = format(dt_now() - timedelta(minutes=1))
        assert dt_min_ago(obj=then) == 1

    def test_min_ago_utc_dt(self):
        """Simple test."""
        then = dt_now() - timedelta(minutes=1)
        assert dt_min_ago(obj=then) == 1

    def test_min_ago_utc_dt_naive(self):
        """Simple test."""
        then = dt_now(None) - timedelta(minutes=1)
        assert dt_min_ago(obj=then) == 1

    def test_min_ago_utc_dtdelta(self):
        """Simple test."""
        then = timedelta(minutes=3)
        assert dt_min_ago(obj=then) == 3

    def test_min_ago_naive(self):
        """Simple test."""
        then = datetime.now() - timedelta(minutes=1)
        assert dt_min_ago(obj=format(then)) == 1


class TestDtNow:
    """Test dt_*."""

    def test_now(self):
        """Pass."""
        now = dt_now()
        assert now.tzinfo

    def test_now_notz(self):
        """Pass."""
        now = dt_now(tz=None)
        assert not now.tzinfo

    def test_now_delta(self):
        """Pass."""
        then = dt_now(delta=timedelta(minutes=5))
        assert dt_min_ago(then) == 5


class TestDtParse:
    """Test dt_*."""

    @pytest.mark.parametrize(
        "val", [format(dt_now()), dt_now(), timedelta(minutes=1)], scope="class",
    )
    def test_val(self, val):
        """Pass."""
        now = dt_parse(obj=val)
        assert isinstance(now, datetime)

    def test_list(self):
        """Pass."""
        now = [format(dt_now())]
        now = dt_parse(obj=now)
        assert isinstance(now, list)
        assert [isinstance(x, datetime) for x in now]


class TestDtWithinMin:
    """Test dt_*."""

    @pytest.mark.parametrize(
        "val", [None, "x", False, True, {}, [], 6, "8", b"9"], scope="class"
    )
    def test_bad(self, val):
        """Pass."""
        then = dt_now(delta=timedelta(minutes=5))
        assert dt_within_min(obj=then, n=val) is False

    @pytest.mark.parametrize("val", [0, 4, "1", b"2"], scope="class")
    def test_ok(self, val):
        """Pass."""
        then = dt_now(delta=timedelta(minutes=5))
        assert dt_within_min(obj=then, n=val) is True



# -*- coding: utf-8 -*-
"""API models for working with adapters and connections."""
from ...constants import AGG_ADAPTER_NAME, AGG_ADAPTER_TITLE, ALL_NAME, NORM_TYPE_MAP
from ...exceptions import ApiError
from ...tools import strip_left, strip_right


def parse_fields(raw):
    """Parse all generic and adapter specific fields.

    Returns:
        :obj:`dict`: parsed generic and adapter specific fields
    """
    parsed = {}
    parsed[AGG_ADAPTER_NAME] = parse_schemas(
        adapter_name=AGG_ADAPTER_NAME,
        adapter_title=AGG_ADAPTER_TITLE,
        adapter_name_raw=f"{AGG_ADAPTER_NAME}_adapter",
        adapter_prefix="specific_data.data",
        all_field="specific_data",
        raw_fields=raw["generic"],
    )

    for raw_name, raw_fields in raw["specific"].items():
        # raw_name = aws_adapter

        prefix = f"adapters_data.{raw_name}"
        # prefix = adapters_data.aws_adapter

        name = strip_right(obj=raw_name, fix="_adapter")
        # name = aws

        title = " ".join(name.split("_")).title()

        fields = parse_schemas(
            adapter_name_raw=raw_name,
            adapter_name=name,
            adapter_prefix=prefix,
            adapter_title=title,
            all_field=prefix,
            raw_fields=raw_fields,
        )
        parsed[name] = fields

    return parsed


def is_complex(field):
    """Determine if a field is complex from its schema."""
    field_type = field["type"]
    field_items_type = field.get("items", {}).get("type")
    if field_type == "array" and field_items_type == "array":
        return True
    return False


def is_root(name, names):
    """Determine if a field is a root field."""
    dots = name.split(".")
    return not (len(dots) > 1 and dots[0] in names)


def parse_complex(field):
    """Pass."""
    field["is_complex"] = is_complex(field=field)
    if field["is_complex"]:
        col_title = field["column_title"]
        col_name = field["column_name"]
        name_base = field["name_base"]
        prefix = field["adapter_prefix"]
        field["sub_fields"] = sub_fields = []
        items = field.pop("items")["items"]

        field_names = [f["name"] for f in items]

        for sub_field in items:
            sub_title = sub_field["title"]
            sub_name = sub_field["name"]
            sub_name_base = f"{name_base}.{sub_name}"
            sub_name_qual = f"{prefix}.{sub_name_base}"

            sub_field["name_base"] = sub_name_base
            sub_field["name_qual"] = sub_name_qual
            sub_field["is_root"] = is_root(name=sub_name, names=field_names)
            sub_field["is_list"] = sub_field["type"] == "array"
            sub_field["parent"] = field["name_qual"]
            sub_field["adapter_name"] = field["adapter_name"]
            sub_field["adapter_name_raw"] = field["adapter_name_raw"]
            sub_field["adapter_prefix"] = field["adapter_prefix"]
            sub_field["adapter_title"] = field["adapter_title"]
            sub_field["column_title"] = f"{col_title}: {sub_title}"
            sub_field["column_name"] = f"{col_name}.{sub_name}"
            sub_field["type_norm"] = get_type_norm(field=sub_field)
            parse_complex(field=sub_field)
            sub_fields.append(sub_field)


def parse_schemas(
    adapter_name_raw, adapter_name, adapter_prefix, adapter_title, all_field, raw_fields,
):
    """Parse field schemas for an adapter."""
    fields = []
    # adapter_name = strip_right(obj=adapter_name_raw, fix="_adapter")
    # adapter_title = " ".join(adapter_name.split("_")).title()

    fields.append(
        {
            "adapter_name_raw": adapter_name_raw,
            "adapter_name": adapter_name,
            "adapter_title": adapter_title,
            "adapter_prefix": adapter_prefix,
            "column_name": f"{adapter_name}:{ALL_NAME}",
            "column_title": f"All {adapter_title} Data",
            "sub_fields": [],
            "is_complex": True,
            "is_list": True,
            "is_root": False,
            "parent": "root",
            "name": ALL_NAME,
            "name_base": ALL_NAME,
            "name_qual": all_field,
            "title": "All Adapter Specific Data",
            "type": "array",
            "type_norm": "complex_complex",
        }
    )

    field_names = [
        strip_left(obj=f["name"], fix=adapter_prefix).strip(".") for f in raw_fields
    ]

    for field in raw_fields:
        title = field["title"]
        name_base = strip_left(obj=field["name"], fix=adapter_prefix).strip(".")
        field["adapter_name"] = adapter_name
        field["adapter_title"] = adapter_title
        field["adapter_prefix"] = adapter_prefix
        field["adapter_name_raw"] = adapter_name_raw
        field["name_base"] = name_base
        field["name_qual"] = field["name"]
        field["is_root"] = is_root(name=name_base, names=field_names)
        field["is_list"] = field["type"] == "array"
        field["parent"] = "root"
        field["column_title"] = f"{adapter_title}: {title}"
        field["column_name"] = f"{adapter_name}:{name_base}"
        field["type_norm"] = get_type_norm(field=field)
        parse_complex(field=field)
        fields.append(field)

    return fields


def get_type_norm(field):
    """Get the normalized type of a field."""
    ftype = field["type"]
    ffmt = field.get("format", "")
    fitype = field.get("items", {}).get("type", "")
    fifmt = field.get("items", {}).get("format", "")
    check = (ftype, ffmt, fitype, fifmt)

    for norm_map, norm_type in NORM_TYPE_MAP:
        if check == norm_map:
            return norm_type

    name = field["name"]
    check = dict(zip(("type", "format", "items.type", "items.format"), check))
    raise ApiError(f"Unmapped normalized type: {name}: {check}")



# -*- coding: utf-8 -*-
"""Test suite."""
import io
import warnings

import pytest
from axonius_api_client.constants import CSV_ADAPTER, DEFAULT_NODE
from axonius_api_client.exceptions import (CnxAddError, CnxGoneError,
                                           CnxTestError, CnxUpdateError,
                                           ConfigInvalidValue, ConfigRequired,
                                           ConfigUnchanged, NotFoundError)
from axonius_api_client.tools import json_dump

from ...meta import CSV_FILECONTENT_STR
from ...utils import get_cnx_broken, get_cnx_existing, get_cnx_working


class TestCnxBase:
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_adapters):
        """Pass."""
        return api_adapters

    @pytest.fixture(scope="class")
    def adapter(self, apiobj):
        """Pass."""
        return apiobj.get_by_name(name=CSV_ADAPTER, node=DEFAULT_NODE)


class TestCnxPrivate(TestCnxBase):
    """Pass."""


class TestCnxPublic(TestCnxBase):
    """Pass."""

    def test_get_by_adapter(self, apiobj):
        """Pass."""
        cnxs = apiobj.cnx.get_by_adapter(
            adapter_name=CSV_ADAPTER, adapter_node=DEFAULT_NODE
        )
        assert isinstance(cnxs, list)
        for cnx in cnxs:
            assert isinstance(cnx, dict)
            assert isinstance(cnx["schemas"], dict)

    def test_get_by_adapter_badname(self, apiobj):
        """Pass."""
        with pytest.raises(NotFoundError):
            apiobj.cnx.get_by_adapter(adapter_name="badwolf", adapter_node=DEFAULT_NODE)

    def test_get_by_adapter_badnode(self, apiobj):
        """Pass."""
        with pytest.raises(NotFoundError):
            apiobj.cnx.get_by_adapter(adapter_name=CSV_ADAPTER, adapter_node="badwolf")

    def test_get_by_uuid(self, apiobj):
        """Pass."""
        cnx = get_cnx_existing(apiobj)
        found = apiobj.cnx.get_by_uuid(
            cnx_uuid=cnx["uuid"],
            adapter_name=cnx["adapter_name"],
            adapter_node=cnx["node_name"],
        )
        assert cnx == found

    def test_get_by_id(self, apiobj):
        """Pass."""
        cnx = get_cnx_existing(apiobj)
        found = apiobj.cnx.get_by_id(
            cnx_id=cnx["id"],
            adapter_name=cnx["adapter_name"],
            adapter_node=cnx["node_name"],
        )
        assert cnx == found

    def test_test_by_id(self, apiobj):
        """Pass."""
        cnx = get_cnx_working(apiobj)
        result = apiobj.cnx.test_by_id(
            cnx_id=cnx["id"],
            adapter_name=cnx["adapter_name"],
            adapter_node=cnx["node_name"],
        )
        assert not result

    def test_test(self, apiobj):
        """Pass."""
        cnx = get_cnx_working(apiobj)
        result = apiobj.cnx.test(
            adapter_name=cnx["adapter_name"],
            adapter_node=cnx["node_name"],
            **cnx["config"],
        )
        assert not result

    def test_test_fail(self, apiobj):
        """Pass."""
        with pytest.raises(CnxTestError):
            apiobj.cnx.test(
                adapter_name="tanium",
                adapter_node=DEFAULT_NODE,
                domain="badwolf",
                username="badwolf",
                password="badwolf",
            )

    def test_test_fail_no_domain(self, apiobj):
        """Pass."""
        with pytest.raises(ConfigRequired):
            apiobj.cnx.test(
                adapter_name="tanium", adapter_node=DEFAULT_NODE, username="x",
            )

    def test_update_cnx_nochange(self, apiobj):
        """Pass."""
        cnx = get_cnx_broken(apiobj)
        with pytest.raises(ConfigUnchanged):
            apiobj.cnx.update_cnx(cnx_update=cnx)

    # XXX broken in pre-3.3
    # @pytest.mark.skip(f"Connection label update not working")
    def test_update_cnx(self, apiobj):
        """Pass."""
        cnx = get_cnx_working(apiobj)
        old_label = cnx["config"].get("connection_label", None)
        gone_test = "gone test"
        if old_label == "badwolf":
            new_label = "badwolf1"
        else:
            new_label = "badwolf"

        cnx_update = apiobj.cnx.update_cnx(cnx_update=cnx, connection_label=new_label)
        assert cnx_update["config"]["connection_label"] == new_label

        cnx_reset = apiobj.cnx.update_by_id(
            cnx_id=cnx_update["id"],
            adapter_name=cnx_update["adapter_name"],
            adapter_node=cnx_update["node_name"],
            connection_label=old_label,
        )
        assert cnx_reset["config"]["connection_label"] == old_label

        with pytest.raises(CnxGoneError):
            apiobj.cnx.update_cnx(cnx_update=cnx, connection_label=gone_test)

        cnx_final = apiobj.cnx.get_by_id(
            cnx_id=cnx_reset["id"], adapter_name=cnx_reset["adapter_name"]
        )
        assert cnx_final["config"]["connection_label"] == old_label

    # XXX broken in pre-3.3
    # @pytest.mark.skip(f"Connection label update not working")
    def test_update_cnx_error(self, apiobj):
        """Pass."""
        cnx = get_cnx_working(apiobj, name="tanium")
        old_label = cnx["config"].get("connection_label", None)
        old_domain = cnx["config"]["domain"]
        new_label = f"old_domain_{old_domain}"
        new_domain = "badwolf"

        if old_domain == new_domain:
            if "old_domain_" in old_label:
                warnings.warn(f"Trying to fix old test data for cnx: {json_dump(cnx)}")
                old_domain = old_label.replace("old_domain_", "")
                cnx = apiobj.cnx.update_cnx(cnx_update=cnx, domain=old_domain)
            else:
                raise Exception(f"Old test data exists for cnx: {json_dump(cnx)}")

        with pytest.raises(CnxUpdateError) as exc:
            apiobj.cnx.update_cnx(
                cnx_update=cnx, connection_label=new_label, domain=new_domain
            )

        assert getattr(exc.value, "cnx_new", None)
        assert getattr(exc.value, "cnx_old", None)
        assert getattr(exc.value, "result", None)
        cnx_update = exc.value.cnx_new
        assert cnx_update["config"]["connection_label"] == new_label
        assert cnx_update["config"]["domain"] == new_domain

        cnx_reset = apiobj.cnx.update_cnx(
            cnx_update=cnx_update, domain=old_domain, connection_label=old_label
        )
        assert cnx_reset["config"]["connection_label"] == old_label
        assert cnx_reset["config"]["domain"] == old_domain

    def test_add_remove(self, apiobj, csv_file_path):
        """Pass."""
        config = {
            "user_id": "badwolf",
            "file_path": csv_file_path,
            # "is_users": False,
            # "is_installed_sw": False,
            # "s3_use_ec2_attached_instance_profile": False,
        }
        cnx_added = apiobj.cnx.add(
            adapter_name=CSV_ADAPTER, adapter_node=DEFAULT_NODE, **config
        )
        for k, v in config.items():
            assert v == cnx_added["config"][k]

        del_result = apiobj.cnx.delete_by_id(
            cnx_id=cnx_added["id"],
            adapter_name=CSV_ADAPTER,
            adapter_node=DEFAULT_NODE,
            delete_entities=True,
        )

        assert not del_result
        # XXX changes pre-3.3
        # assert del_result["client_id"] == cnx_added["id"]

        with pytest.raises(NotFoundError):
            apiobj.cnx.get_by_id(
                cnx_id=cnx_added["id"],
                adapter_name=CSV_ADAPTER,
                adapter_node=DEFAULT_NODE,
            )

    def test_add_remove_str_error(self, apiobj):
        """Pass."""
        config = {
            "user_id": "badwolf",
            "file_path": io.StringIO("BADwoLF"),
        }
        with pytest.raises(CnxAddError) as exc:
            apiobj.cnx.add(adapter_name=CSV_ADAPTER, adapter_node=DEFAULT_NODE, **config)

        assert getattr(exc.value, "cnx_new", None)
        assert not hasattr(exc.value, "cnx_old")
        assert getattr(exc.value, "result", None)
        cnx_added = exc.value.cnx_new

        del_result = apiobj.cnx.delete_by_id(
            cnx_id=cnx_added["id"],
            adapter_name=CSV_ADAPTER,
            adapter_node=DEFAULT_NODE,
            delete_entities=True,
        )

        assert not del_result
        # XXX changes pre-3.3
        # assert del_result["client_id"] == cnx_added["id"]

        with pytest.raises(NotFoundError):
            apiobj.cnx.get_by_id(
                cnx_id=cnx_added["id"],
                adapter_name=CSV_ADAPTER,
                adapter_node=DEFAULT_NODE,
                retry=2,
            )

    def test_add_remove_path(self, apiobj, tmp_path):
        """Pass."""
        file_path = tmp_path / "test.csv"
        file_path.write_text(CSV_FILECONTENT_STR)
        config = {
            "user_id": "badwolf",
            "file_path": file_path,
        }
        cnx_added = apiobj.cnx.add(
            adapter_name=CSV_ADAPTER, adapter_node=DEFAULT_NODE, **config
        )

        assert isinstance(cnx_added["config"]["file_path"], dict)

        del_result = apiobj.cnx.delete_by_id(
            cnx_id=cnx_added["id"],
            adapter_name=CSV_ADAPTER,
            adapter_node=DEFAULT_NODE,
            delete_entities=True,
        )

        assert not del_result
        # XXX changes pre-3.3
        # assert del_result["client_id"] == cnx_added["id"]

        with pytest.raises(NotFoundError):
            apiobj.cnx.get_by_id(
                cnx_id=cnx_added["id"],
                adapter_name=CSV_ADAPTER,
                adapter_node=DEFAULT_NODE,
            )

    def test_add_remove_path_notexists(self, apiobj, tmp_path):
        """Pass."""
        file_path = tmp_path / "badtest.csv"
        config = {
            "user_id": "badwolf",
            "file_path": file_path,
        }
        with pytest.raises(ConfigInvalidValue):
            apiobj.cnx.add(adapter_name=CSV_ADAPTER, adapter_node=DEFAULT_NODE, **config)

    def test_add_remove_stream(self, apiobj):
        """Pass."""
        config = {
            "user_id": "badwolf",
            "file_path": io.StringIO(CSV_FILECONTENT_STR),
        }
        cnx_added = apiobj.cnx.add(
            adapter_name=CSV_ADAPTER, adapter_node=DEFAULT_NODE, **config
        )

        assert isinstance(cnx_added["config"]["file_path"], dict)

        del_result = apiobj.cnx.delete_by_id(
            cnx_id=cnx_added["id"],
            adapter_name=CSV_ADAPTER,
            adapter_node=DEFAULT_NODE,
            delete_entities=True,
        )

        assert not del_result
        # XXX changes pre-3.3
        # assert del_result["client_id"] == cnx_added["id"]

        with pytest.raises(NotFoundError):
            apiobj.cnx.get_by_id(
                cnx_id=cnx_added["id"],
                adapter_name=CSV_ADAPTER,
                adapter_node=DEFAULT_NODE,
            )

    def test_add_remove_error(self, apiobj, csv_file_path_broken):
        """Pass."""
        config = {
            "user_id": "badwolf",
            "file_path": csv_file_path_broken,
            # SANE_DEFAULTS handles these now:
            # "is_users": False,
            # "is_installed_sw": False,
            # "s3_use_ec2_attached_instance_profile": False,
        }
        with pytest.raises(CnxAddError) as exc:
            apiobj.cnx.add(adapter_name=CSV_ADAPTER, adapter_node=DEFAULT_NODE, **config)

        assert getattr(exc.value, "cnx_new", None)
        assert not hasattr(exc.value, "cnx_old")
        assert getattr(exc.value, "result", None)
        cnx_added = exc.value.cnx_new

        del_result = apiobj.cnx.delete_by_id(
            cnx_id=cnx_added["id"],
            adapter_name=CSV_ADAPTER,
            adapter_node=DEFAULT_NODE,
            delete_entities=True,
        )

        assert not del_result
        # XXX changes pre-3.3
        # assert del_result["client_id"] == cnx_added["id"]

        with pytest.raises(NotFoundError):
            apiobj.cnx.get_by_id(
                cnx_id=cnx_added["id"],
                adapter_name=CSV_ADAPTER,
                adapter_node=DEFAULT_NODE,
                retry=2,
            )



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
import re

from ...api.parsers.tables import tablize
from ...tools import json_dump
from ..context import CONTEXT_SETTINGS, click
from ..options import AUTH, add_options

OPTIONS = [
    *AUTH,
    click.option(
        "--adapter-re",
        "-ar",
        "adapter_re",
        default=".*",
        help="Only fetch fields for adapters matching this regex",
        metavar="REGEX",
        show_envvar=True,
        show_default=True,
    ),
    click.option(
        "--field-re",
        "-fr",
        "field_re",
        default=".*",
        help="Only fetch fields matching this regex",
        metavar="REGEX",
        show_envvar=True,
        show_default=True,
    ),
    click.option(
        "--field-key",
        "-fk",
        "field_key",
        default="name_qual",
        help="Which field key to match against for --field-re",
        type=click.Choice(
            ["name_qual", "name", "name_base", "column_title", "column_name"]
        ),
        show_envvar=True,
        show_default=True,
    ),
    click.option(
        "--export-format",
        "-xf",
        "export_format",
        type=click.Choice(["json-full", "json", "str", "table"]),
        help="Control how much schema information to return",
        default="str",
        show_envvar=True,
        show_default=True,
    ),
]


@click.command(name="get-fields", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(
    ctx,
    url,
    key,
    secret,
    adapter_re,
    field_re,
    field_key,
    export_format,
    help_detailed=None,
    **kwargs
):
    """Get the available fields (columns) for assets."""
    p_grp = ctx.parent.command.name

    client = ctx.obj.start_client(url=url, key=key, secret=secret)
    apiobj = getattr(client, p_grp)

    are = re.compile(adapter_re, re.I)
    fre = re.compile(field_re, re.I)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        rows = apiobj.fields.get()

    matches = []

    for adapter, schemas in rows.items():
        if not are.search(adapter):
            continue

        for schema in schemas:
            if not fre.search(schema[field_key]):
                continue

            matches.append(schema)

    if export_format == "json-full":
        click.secho(json_dump(matches))
        ctx.exit(0)

    if export_format == "json":
        keys = {
            "adapter_name": "Adapter",
            "column_name": "Name",
            "column_title": "Title",
            "is_complex": "Is complex",
            "is_root": "Is root",
            "type_norm": "Normalized Type",
            "name_qual": "Fully Qualified Name",
        }
        matches = [{k: v for k, v in x.items() if k in keys} for x in matches]
        click.secho(json_dump(matches))
        ctx.exit(0)

    if export_format == "str":
        for x in matches:
            click.secho(x["column_name"])
        ctx.exit(0)

    if export_format == "table":
        keys = {
            "adapter_name": "Adapter",
            "column_name": "Name",
            "column_title": "Title",
            "type_norm": "Normalized Type",
        }
        matches = [{keys[k]: v for k, v in x.items() if k in keys} for x in matches]
        click.secho(tablize(value=matches, err=None, fmt="simple", footer=True))
        ctx.exit(0)



# -*- coding: utf-8 -*-
"""Authentication methods."""
import abc

from .api.routers import API_VERSION
from .constants import LOG_LEVEL_AUTH
from .exceptions import AlreadyLoggedIn, AuthError, InvalidCredentials, NotLoggedIn
from .logs import get_obj_log


class Model:
    """Abstract base class for all Authentication methods."""

    @abc.abstractmethod
    def login(self):
        """Login to API."""
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def logout(self):
        """Logout from API."""
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def check_login(self):
        """Throw exc if not login."""
        raise NotImplementedError  # pragma: no cover

    @abc.abstractproperty
    def http(self):
        """Get HttpClient object."""
        raise NotImplementedError  # pragma: no cover

    @abc.abstractproperty
    def is_logged_in(self):
        """Check if login has been called."""
        raise NotImplementedError  # pragma: no cover


class Mixins:
    """Mixins for Model."""

    _logged_in = False
    """:obj:`bool`: Attribute checked by :meth:`is_logged_in`."""

    def __init__(self, http, creds, **kwargs):
        """Mixins for Model.

        Args:
            http (:obj:`.http.Http`):
                HTTP client to use to send requests.
            creds:
                Credentials used by this Auth method.

        """
        log_level = kwargs.get("log_level", LOG_LEVEL_AUTH)
        self.LOG = get_obj_log(obj=self, level=log_level)
        """:obj:`logging.Logger`: Logger for this object."""

        self._http = http
        """:obj:`.http.Http`: HTTP Client."""

        self._creds = creds
        """:obj:`dict`: Credential store."""

        self._check_http_lock()
        self._set_http_lock()

    def __str__(self):
        """Show object info.

        Returns:
            :obj:`str`

        """
        bits = [f"url={self.http.url!r}", f"is_logged_in={self.is_logged_in}"]
        bits = ", ".join(bits)
        return f"{self.__class__.__module__}.{self.__class__.__name__}({bits})"

    def __repr__(self):
        """Show object info.

        Returns:
            :obj:`str`

        """
        return self.__str__()

    @property
    def http(self):
        """Get HttpClient object.

        Returns:
            :obj:`.http.Http`

        """
        return self._http

    def _check_http_lock(self):
        """Check HTTP client not already used by another Auth.

        Raises:
            :exc:`AuthError`

        """
        auth_lock = getattr(self.http, "_auth_lock", None)
        if auth_lock:
            raise AuthError(f"{self.http} already being used by {auth_lock}")

    def _set_http_lock(self):
        """Set HTTP Client auth lock."""
        self._http._auth_lock = self

    def _validate(self):
        """Validate credentials."""
        response = self.http(method="get", path=API_VERSION.devices.count)

        try:
            response.raise_for_status()
        except Exception as exc:
            self._logged_in = False
            raise InvalidCredentials(
                f"Invalid credentials on {self} -- exception: {exc}"
            )

        self._logged_in = True

    def logout(self):
        """Logout from API."""
        self.check_login()
        self._logged_in = False
        self._logout()

    def check_login(self):
        """Throw exc if not login.

        Raises:
            :exc:`NotLoggedIn`

        """
        if not self.is_logged_in:
            raise NotLoggedIn(f"Must call login() on {self}")

    @property
    def is_logged_in(self):
        """Check if login has been called.

        Returns:
            :obj:`bool`

        """
        return self._logged_in


class ApiKey(Mixins, Model):
    """Authentication method using API key & API secret."""

    def __init__(self, http, key, secret, **kwargs):
        """Authenticate using API key & API secret.

        Args:
            http (:obj:`.http.Http`):
                HTTP client to use to send requests.
            key (:obj:`str`):
                API key to use in credentials.
            secret (:obj:`str`):
                API secret to use in credentials.

        """
        creds = {"key": key, "secret": secret}
        super(ApiKey, self).__init__(http=http, creds=creds, **kwargs)

    @property
    def _cred_fields(self):
        return ["key", "secret"]

    def _logout(self):
        """Logout from API."""
        self._logged_in = False
        self.http.session.headers = {}

    def login(self):
        """Login to API."""
        if self.is_logged_in:
            raise AlreadyLoggedIn(f"Already logged in on {self}")

        self.http.session.headers["api-key"] = self._creds["key"]
        self.http.session.headers["api-secret"] = self._creds["secret"]
        self._validate()
        self._logged_in = True
        self.LOG.debug(f"Successfully logged in using {self._cred_fields}")



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ...context import CONTEXT_SETTINGS, click
from ...options import AUTH, NODE_CNX, add_options
from .grp_common import ID_CNX, PROMPTS

OPTIONS = [*AUTH, *NODE_CNX, *PROMPTS, ID_CNX]


@click.command(name="test-by-id", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, adapter_name, adapter_node, cnx_id, **kwargs):
    """Test reachability for an existing connection by ID."""
    client = ctx.obj.start_client(url=url, key=key, secret=secret)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        client.adapters.cnx.test_by_id(
            adapter_name=adapter_name, adapter_node=adapter_node, cnx_id=cnx_id,
        )
        ctx.obj.echo_ok(msg=f"Reachability test succeeded!")



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
import click

from ..context import AliasedGroup
from . import (cmd_config_get, cmd_config_update, cmd_config_update_from_json,
               cmd_file_upload, cmd_get, grp_cnx)


@click.group(cls=AliasedGroup)
def adapters():
    """Group: Work with adapters and adapter connections."""


adapters.add_command(cmd_get.cmd)
adapters.add_command(cmd_config_get.cmd)
adapters.add_command(cmd_config_update.cmd)
adapters.add_command(cmd_config_update_from_json.cmd)
adapters.add_command(cmd_file_upload.cmd)
adapters.add_command(grp_cnx.cnx)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
import os

import click
import dotenv

from ...tools import get_path
from ..options import AUTH, add_options


@click.command(name="write-config")
@add_options(AUTH)
@click.pass_context
def cmd(ctx, url, key, secret):
    """Create/Update a '.env' file with url, key, and secret.

    File is created in the current working directory.
    """
    ctx.obj.start_client(url=url, key=key, secret=secret)

    cwd = os.getcwd()
    path = get_path(cwd) / ".env"

    if not path.is_file():
        msg = "Creating file {!r}".format(format(path))
        click.secho(message=msg, err=True, fg="green")

        path.touch()
        path.chmod(0o600)

    path = format(path)
    msg = "Setting AX_URL, AX_KEY, and AX_SECRET in {!r}".format(format(path))
    click.secho(message=msg, err=True, fg="green")
    dotenv.set_key(dotenv_path=path, key_to_set="AX_URL", value_to_set=url)
    dotenv.set_key(dotenv_path=path, key_to_set="AX_KEY", value_to_set=key)
    dotenv.set_key(dotenv_path=path, key_to_set="AX_SECRET", value_to_set=secret)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
import click

from ..context import AliasedGroup
from . import (cmd_convert_cert, cmd_shell, cmd_signup, cmd_sysinfo,
               cmd_write_config)


@click.group(cls=AliasedGroup)
def tools():
    """Group: CLI tools."""


tools.add_command(cmd_shell.cmd)
tools.add_command(cmd_write_config.cmd)
tools.add_command(cmd_sysinfo.cmd)
tools.add_command(cmd_signup.cmd)
tools.add_command(cmd_convert_cert.cmd)



# -*- coding: utf-8 -*-
"""API model for working with system configuration."""
from ...constants import DEFAULT_PERM, VALID_PERMS
from ...exceptions import ApiError, NotFoundError
from ...tools import listify
from ..mixins import ChildMixins

ROLE_ACTIONS = [
    "delete",
    "post",
    "put",
    "run",
    "run_manual_discovery",
    "reset_api_key",
    "get_users_and_roles",
    "get",
    "post",
    "put",
    "run",
]


def parse_roles_labels(raw, role=None):
    """Pass."""
    # raw = {k.split(".")[1:]: v for k, v in raw.items()}
    # permissions = (role or {}).get("permissions", {})
    parsed = {}
    # parsed = {
    #     k[0]: {"actions": {}, "description": v, "parent": None}
    #     for k, v in raw.items()
    #     if len(k) == 1
    # }

    for perm, desc in raw.items():
        items = perm.split(".")[1:]
        category = items.pop(0)
        if not items:
            print(category)
            parsed[category] = {"actions": {}, "description": desc, "parent": None}

    # for perm, description in raw.items():
    #     items = perm.split(".")[1:]
    #     category = items.pop(0)
    #     category_perms = permissions.get(category, {})
    #     parsed[category] = parsed.get(
    #         category, {"actions": {}, "description": None, "parent": None}
    #     )

    #     if not items:
    #         parsed[category]["description"] = description
    #         continue

    #     action = items.pop(0)

    #     if action in ROLE_ACTIONS:
    #         parsed[category]["actions"][action] = {
    #             "description": description,
    #             "value": category_perms.get(action, False),
    #         }
    #         continue

    #     section = action
    #     section_perms = category_perms.get(section, {})

    #     parsed[section] = parsed.get(
    #         section,
    #         {
    #             "actions": {},
    #             "parent": category,
    #             "description": f"Section under {category}",
    #         },
    #     )

    #     if items:
    #         section_action = items.pop(0)

    #         if section_action in ROLE_ACTIONS:
    #             parsed[section]["actions"] = {
    #                 "description": description,
    #                 "value": section_perms.get(section_action, False),
    #             }

    return parsed


class Roles(ChildMixins):
    """Child API object to work with Roles."""

    def get_categories(self, role=None):
        """Pass."""
        labels = self._get_labels()
        return parse_roles_labels(raw=labels, role=role)

    def get(self):
        """Pass."""
        roles = self._get()
        return roles

    def get_by_name(self, name):
        """Pass."""
        roles = self.get()

        valid = [x["name"] for x in roles]

        if name not in valid:
            valid = "\n" + "\n".join(valid)
            raise NotFoundError(f"Role name {name!r} not found, valid roles:{valid}")

        return [x for x in roles if x["name"] == name][0]

    def get_by_uuid(self, uuid):
        """Pass."""
        roles = self.get()

        valid = [x["uuid"] for x in roles]

        if uuid not in valid:
            valid = "\n" + "\n".join(valid)
            raise NotFoundError(f"Role uuid {uuid!r} not found, valid roles:{valid}")

        return [x for x in roles if x["uuid"] == uuid][0]

    def validate_perms(self, **kwargs):
        """Pass."""
        perms = self.get_categories()

        for category, actions in kwargs.items():
            if category not in perms:
                valid = "\n" + "\n".join(list(perms))
                raise NotFoundError(f"Category {category!r} not found, valid: {valid}")
            actions = listify(actions)

    def add(self, name, **kwargs):
        """Pass."""
        roles = self.get()
        names = [x["user_name"] for x in roles]

        if name in names:
            raise ApiError(f"Role named {name!r} already exists")

        permissions = {}

        self._add(name=name, permissions=permissions)
        return self.get(name=name)

    def update(
        self,
        name,
        adapters=DEFAULT_PERM,
        dashboard=DEFAULT_PERM,
        devices=DEFAULT_PERM,
        enforcements=DEFAULT_PERM,
        instances=DEFAULT_PERM,
        reports=DEFAULT_PERM,
        settings=DEFAULT_PERM,
        users=DEFAULT_PERM,
    ):
        """Pass."""
        # role = self.get_by_name(name=name)

        permissions = self._check_valid_perms(
            adapters=adapters,
            dashboard=dashboard,
            devices=devices,
            enforcements=enforcements,
            instances=instances,
            reports=reports,
            settings=settings,
            users=users,
        )
        self._update(name=name, permissions=permissions)
        return self.get(name=name)

    # XXX predefined
    def delete(self, name):
        """Pass."""
        role = self.get_by_name(name=name)

        if role.get("predefined"):
            raise ApiError(f"Not allowed to delete predefined role {name!r}")

        return self._delete(uuid=role["uuid"])

    def _get(self):
        """Direct API method to get known roles.

        Returns:
            :obj:`list` of :obj:`str`: known roles
        """
        path = self.router.roles
        return self.request(method="get", path=path)

    def _add(self, name, permissions):
        """Direct API method to add a role.

        Args:
            name (:obj:`str`): name of new role
            permissions (:obj:`dict`): permissions for new role
        """
        data = {"name": name, "permissions": permissions}
        path = self.router.roles
        return self.request(method="put", path=path, json=data)

    def _update(self, name, permissions):
        """Direct API method to update a roles permissions.

        Args:
            name (:obj:`str`): name of role to update
            permissions (:obj:`dict`): permissions to update on new role
        """
        data = {"name": name, "permissions": permissions}
        path = self.router.roles
        return self.request(
            method="post", path=path, json=data, error_json_invalid=False
        )

    def _delete(self, uuid):
        """Direct API method to delete a role.

        Args:
            name (:obj:`str`): name of role to delete
        """
        path = self.router.roles_by_uuid.format(uuid=uuid)
        return self.request(method="delete", path=path, error_json_invalid=False)

    def _check_valid_perm(self, name, value):
        """Pass."""
        if value not in VALID_PERMS:
            valids = "\n" + "\n".join(VALID_PERMS)
            msg = f"Invalid permission {value!r} for {name!r} - Must be one of:{valids}"
            raise ApiError(msg)

    def _check_valid_perms(
        self,
        adapters,
        dashboard,
        devices,
        enforcements,
        instances,
        reports,
        settings,
        users,
    ):
        """Pass."""
        self._check_valid_perm(name="adapters", value=adapters)
        self._check_valid_perm(name="dashboard", value=dashboard)
        self._check_valid_perm(name="devices", value=devices)
        self._check_valid_perm(name="enforcements", value=enforcements)
        self._check_valid_perm(name="instances", value=instances)
        self._check_valid_perm(name="reports", value=reports)
        self._check_valid_perm(name="settings", value=settings)
        self._check_valid_perm(name="users", value=users)

        return {
            "Adapters": adapters,
            "Dashboard": dashboard,
            "Devices": devices,
            "Enforcements": enforcements,
            "Instances": instances,
            "Reports": reports,
            "Settings": settings,
            "Users": users,
        }

    def _get_labels(self):
        """Pass."""
        path = self.router.roles_labels
        return self.request(method="get", path=path)



# -*- coding: utf-8 -*-
"""Test suite for assets."""
import copy

import pytest

from .callbacks import Callbacks, load_test_data


class CallbacksBase(Callbacks):
    """Pass."""

    @pytest.fixture(scope="class")
    def cbexport(self):
        """Pass."""
        return "base"

    def test_row_as_is(self, cbexport, apiobj, caplog):
        """Pass."""
        getargs = {}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)
        cbobj.start()

        rows = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][:5])

        for row in rows:
            rows_ret = cbobj.row(row=copy.deepcopy(row))
            for row_ret in rows_ret:
                assert row == row_ret
        cbobj.stop()

    def test_row_fully_loaded(self, cbexport, apiobj, caplog):
        """Pass."""
        getargs = {
            "field_excludes": ["adapters"],
            "field_flatten": True,
            "field_titles": True,
            "field_join": True,
            "field_null": True,
            "report_adapters_missing": True,
        }
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)
        cbobj.start()

        rows = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][:5])

        for row in rows:
            rows_ret = cbobj.row(row=copy.deepcopy(row))
            for row_ret in rows_ret:
                assert row != row_ret

        cbobj.stop()


class TestDevicesCallbacksBase(CallbacksBase):
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_devices):
        """Pass."""
        return load_test_data(apiobj=api_devices)


class TestUsersCallbacksBase(CallbacksBase):
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_users):
        """Pass."""
        return load_test_data(apiobj=api_users)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ....tools import json_dump
from ...context import CONTEXT_SETTINGS, click
from ...options import AUTH, add_options

EXPORT = click.option(
    "--export-format",
    "-xf",
    "export_format",
    type=click.Choice(["json", "str"]),
    help="Format of to export data in",
    default="str",
    show_envvar=True,
    show_default=True,
)

OPTIONS = [
    *AUTH,
    EXPORT,
]


@click.command(name="about", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, export_format, **kwargs):
    """Get instance version information."""
    client = ctx.obj.start_client(url=url, key=key, secret=secret)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        data = client.system.meta.about()

    if export_format == "json":
        click.secho(json_dump(data))
        ctx.exit(0)

    if export_format == "str":
        for k, v in data.items():
            k = k.replace("_", " ").title()
            click.secho(f"{k}: {v}")
        ctx.exit(0)

    ctx.exit(1)



# -*- coding: utf-8 -*-
"""Test suite for axonius_api_client.tools."""
from ....cli import cli
from ....tools import json_load
from ...utils import load_clirunner


class TestGrpAdaptersCmdGet:
    """Pass."""

    def test_json_full(self, request, monkeypatch):
        """Pass."""
        runner = load_clirunner(request, monkeypatch)

        args1 = ["adapters", "get", "--export-format", "json-full"]
        result1 = runner.invoke(cli=cli, args=args1)

        stderr1 = result1.stderr
        stdout1 = result1.stdout
        exit_code1 = result1.exit_code

        assert stdout1
        assert stderr1
        assert exit_code1 == 0

        json1 = json_load(stdout1)
        assert isinstance(json1, list)

    def test_json_basic(self, request, monkeypatch):
        """Pass."""
        runner = load_clirunner(request, monkeypatch)

        args1 = ["adapters", "get", "--export-format", "json"]
        result1 = runner.invoke(cli=cli, args=args1)

        stderr1 = result1.stderr
        stdout1 = result1.stdout
        exit_code1 = result1.exit_code

        assert stdout1
        assert stderr1
        assert exit_code1 == 0

        json1 = json_load(stdout1)
        assert isinstance(json1, list)

    def test_table(self, request, monkeypatch):
        """Pass."""
        runner = load_clirunner(request, monkeypatch)

        args1 = ["adapters", "get", "--export-format", "table"]
        result1 = runner.invoke(cli=cli, args=args1)

        stderr1 = result1.stderr
        stdout1 = result1.stdout
        exit_code1 = result1.exit_code

        assert stdout1
        assert stderr1
        assert exit_code1 == 0

    def test_str(self, request, monkeypatch):
        """Pass."""
        runner = load_clirunner(request, monkeypatch)

        args1 = ["adapters", "get", "--export-format", "str"]
        result1 = runner.invoke(cli=cli, args=args1)

        stderr1 = result1.stderr
        stdout1 = result1.stdout
        exit_code1 = result1.exit_code

        assert stdout1
        assert stderr1
        assert exit_code1 == 0

    def test_str_args(self, request, monkeypatch):
        """Pass."""
        runner = load_clirunner(request, monkeypatch)

        args1 = ["adapters", "get", "--export-format", "str-args"]
        result1 = runner.invoke(cli=cli, args=args1)

        stderr1 = result1.stderr
        stdout1 = result1.stdout
        exit_code1 = result1.exit_code

        assert stdout1
        assert stderr1
        assert exit_code1 == 0



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ..context import CONTEXT_SETTINGS, click
from ..options import AUTH, add_options

OPTIONS = [*AUTH]


@click.command(name="get-fields-default", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, **kwargs):
    """Get the default fields (columns) for assets."""
    p_grp = ctx.parent.command.name

    client = ctx.obj.start_client(url=url, key=key, secret=secret)
    apiobj = getattr(client, p_grp)
    rows = apiobj.fields_default

    for row in rows:
        click.secho(row)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
import click

from ...tools import json_dump, sysinfo
from ..options import add_options

EXPORT = click.option(
    "--export-format",
    "-xf",
    "export_format",
    type=click.Choice(["json", "str"]),
    help="Format of to export data in",
    default="str",
    show_envvar=True,
    show_default=True,
)

OPTIONS = [EXPORT]


@click.command(name="sysinfo")
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, export_format):
    """Print out system and python information."""
    data = sysinfo()

    if export_format == "str":
        for k, v in data.items():
            click.secho(f"{k}: {v}")
        ctx.exit(0)

    if export_format == "json":
        click.secho(json_dump(data))
        ctx.exit(0)

    ctx.exit(1)



# -*- coding: utf-8 -*-
"""API model for working with system configuration."""
from ..mixins import ChildMixins


class Nodes(ChildMixins):
    """Child API model for working with instances."""

    def get(self):
        """Get instances.

        Returns:
            :obj:`dict`: instances
        """
        return self._get()["instances"]

    # XXX add get_by_name, get_core, get_collectors

    def _init(self, parent):
        """Post init method for subclasses to use for extra setup.

        Args:
            parent (:obj:`.api.mixins.Model`): parent API model of this child
        """
        super(Nodes, self)._init(parent=parent)

    def _get(self):
        """Direct API method to get instances.

        Returns:
            :obj:`dict`: instances
        """
        """Example return
        {
            "connection_data": {
                "host": "<axonius-hostname>",
                "key": "QV2bu53oV3HMGugMOvgNSa1B5dNCDCY0"
            },
            "instances": [
                {
                    "hostname": "builds-vm-jim-2-15-b-1581114965-000",
                    "ips": [
                        "10.20.0.100"
                    ],
                    "last_seen": "Sun, 09 Feb 2020 22:19:22 GMT",
                    "node_id": "a7afe3af5d05428dbecc55704fc7e3ea",
                    "node_name": "Master",
                    "node_user_password": "",
                    "status": "Activated",
                    "tags": {}
                }
            ]
        }
        """
        return self.request(method="get", path=self.router.instances)

    def _delete(self, node_id):  # pragma: no cover
        """Pass."""
        data = {"nodeIds": node_id}
        path = self.router.instances
        return self.request(method="delete", path=path, json=data)

    def _update(self, node_id, node_name, hostname):  # pragma: no cover
        """Direct API method to update an instance.

        Args:
            node_id (:obj:`str`): node id of instance
            node_name (:obj:`str`): node name of instance
            hostname (:obj:`str`): hostname of instance

        Returns:
            :obj:`dict`: updated instance
        """
        """Example return

        {
            "nodeIds": "a7afe3af5d05428dbecc55704fc7e3ea",
            "node_name": "Masterx",
            "hostname": "builds-vm-jim-2-15-b-1581114965-000"
        }
        """
        data = {"nodeIds": node_id, "node_name": node_name, "hostname": hostname}
        path = self.router.instances
        return self.request(method="update", path=path, json=data)



# -*- coding: utf-8 -*-
"""API models package."""
from . import asset_mixin, devices, fields, labels, saved_query, users
from .devices import Devices
from .users import Users

__all__ = (
    "Users",
    "Devices",
    "users",
    "devices",
    "fields",
    "asset_mixin",
    "labels",
    "saved_query",
)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
import click

from .. import version
from ..constants import (LOG_FILE_MAX_FILES, LOG_FILE_MAX_MB, LOG_FILE_NAME,
                         LOG_FILE_PATH, LOG_LEVEL_API, LOG_LEVEL_AUTH,
                         LOG_LEVEL_CONSOLE, LOG_LEVEL_FILE, LOG_LEVEL_HTTP,
                         LOG_LEVEL_PACKAGE, LOG_LEVELS_STR, REQUEST_ATTR_MAP,
                         RESPONSE_ATTR_MAP, TIMEOUT_CONNECT, TIMEOUT_RESPONSE)
from . import context, grp_adapters, grp_assets, grp_system, grp_tools


@click.group(
    cls=context.AliasedGroup,
    context_settings=context.CONTEXT_SETTINGS,
    epilog="""
All of the options listed above must be supplied BEFORE any commands or groups.
""",
)
@click.option(
    "--quiet/--no-quiet",
    "-q/-nq",
    "quiet",
    default=False,
    help="Silence green text.",
    show_envvar=True,
    show_default=True,
)
@click.option(
    "--log-level-package",
    "-lvlpkg",
    "log_level_package",
    default=LOG_LEVEL_PACKAGE,
    help="Logging level to use for entire package.",
    type=click.Choice(LOG_LEVELS_STR),
    show_envvar=True,
    show_default=True,
)
@click.option(
    "--log-level-http",
    "-lvlhttp",
    "log_level_http",
    default=LOG_LEVEL_HTTP,
    help="Logging level to use for http client.",
    type=click.Choice(LOG_LEVELS_STR),
    show_envvar=True,
    show_default=True,
)
@click.option(
    "--log-level-auth",
    "-lvlauth",
    "log_level_auth",
    default=LOG_LEVEL_AUTH,
    help="Logging level to use for auth client.",
    type=click.Choice(LOG_LEVELS_STR),
    show_envvar=True,
    show_default=True,
)
@click.option(
    "--log-level-api",
    "-lvlapi",
    "log_level_api",
    default=LOG_LEVEL_API,
    help="Logging level to use for api clients.",
    type=click.Choice(LOG_LEVELS_STR),
    show_envvar=True,
    show_default=True,
)
@click.option(
    "--log-level-console",
    "-lvlcon",
    "log_level_console",
    default=LOG_LEVEL_CONSOLE,
    help="Logging level to use for console output.",
    type=click.Choice(LOG_LEVELS_STR),
    show_envvar=True,
    show_default=True,
)
@click.option(
    "--log-level-file",
    "-lvlfile",
    "log_level_file",
    default=LOG_LEVEL_FILE,
    help="Logging level to use for file output.",
    type=click.Choice(LOG_LEVELS_STR),
    show_envvar=True,
    show_default=True,
)
@click.option(
    "--log-request-attrs",
    "-reqattr",
    "log_request_attrs",
    help="Log http client request attributes.",
    multiple=True,
    type=click.Choice(list(REQUEST_ATTR_MAP) + ["all"]),
    show_envvar=True,
)
@click.option(
    "--log-response-attrs",
    "-respattr",
    "log_response_attrs",
    help="Log http client response attributes.",
    multiple=True,
    type=click.Choice(list(RESPONSE_ATTR_MAP) + ["all"]),
    show_envvar=True,
)
@click.option(
    "--log-request-body",
    "-reqbody",
    "log_request_body",
    default=False,
    help="Log http client request body.",
    is_flag=True,
    show_envvar=True,
)
@click.option(
    "--log-response-body",
    "-respbody",
    "log_response_body",
    help="Log http client response body.",
    default=False,
    is_flag=True,
    show_envvar=True,
)
@click.option(
    "--log-console",
    "-c",
    "log_console",
    default=False,
    help="Enable logging to STDERR.",
    is_flag=True,
    show_envvar=True,
)
@click.option(
    "--log-file",
    "-f",
    "log_file",
    default=False,
    help="Enable logging to -fn/--log-file-name in -fp/--log-file-path.",
    is_flag=True,
    show_envvar=True,
)
@click.option(
    "--log-file-name",
    "-fn",
    "log_file_name",
    metavar="FILENAME",
    default=LOG_FILE_NAME,
    help="Log file to save logs to if -f/--log-file supplied.",
    show_envvar=True,
    show_default=True,
)
@click.option(
    "--log-file-path",
    "-fp",
    "log_file_path",
    metavar="PATH",
    default=LOG_FILE_PATH,
    help="Directory to use for -fn/--log-file-name (Defaults to CWD).",
    show_envvar=True,
)
@click.option(
    "--log-file-max-mb",
    "-fmb",
    "log_file_max_mb",
    default=LOG_FILE_MAX_MB,
    help="Rollover -fn/--log-file-name at this many megabytes.",
    type=click.INT,
    show_envvar=True,
    show_default=True,
)
@click.option(
    "--log-file-max-files",
    "-fmf",
    "log_file_max_files",
    default=LOG_FILE_MAX_FILES,
    help="Keep this many rollover logs.",
    type=click.INT,
    show_envvar=True,
    show_default=True,
)
@click.option(
    "--proxy",
    "-p",
    "proxy",
    default="",
    help="Proxy to use to connect to Axonius.",
    metavar="PROXY",
    show_envvar=True,
    show_default=True,
)
@click.option(
    "--cert-client-both",
    "-ccb",
    "cert_client_both",
    type=click.Path(exists=True, resolve_path=True),
    help="Path to client SSL certificate and private key in one file for mutual TLS.",
    metavar="PATH",
    show_envvar=True,
    show_default=True,
)
@click.option(
    "--cert-client-cert",
    "-ccc",
    "cert_client_cert",
    type=click.Path(exists=True, resolve_path=True),
    help="Path to client SSL certificate for mutual TLS.",
    metavar="PATH",
    show_envvar=True,
    show_default=True,
)
@click.option(
    "--cert-client-key",
    "-cck",
    "cert_client_key",
    type=click.Path(exists=True, resolve_path=True),
    help="Path to client SSL private key for mutual TLS",
    metavar="PATH",
    show_envvar=True,
    show_default=True,
)
@click.option(
    "--certpath",
    "-cp",
    "certpath",
    type=click.Path(exists=True, resolve_path=True),
    help="Path to SSL certificate for verifying the certificate offered by Axonius.",
    metavar="PATH",
    show_envvar=True,
    show_default=True,
)
@click.option(
    "--certverify",
    "-cv",
    "certverify",
    default=False,
    help=(
        "Perform SSL Certificate Verification (will fail if cert is self-signed"
        " or not signed by a system CA)."
    ),
    is_flag=True,
    show_envvar=True,
)
@click.option(
    "--no-certwarn",
    "-ncw",
    "certwarn",
    default=True,
    help="Disable warnings for self-signed SSL certificates.",
    is_flag=True,
    show_envvar=True,
)
@click.option(
    "--no-wraperror",
    "-nw",
    "wraperror",
    default=True,
    help="Show the full traceback of exceptions instead of a wrapped error.",
    is_flag=True,
    show_envvar=True,
)
@click.option(
    "--timeout-connect",
    "-tc",
    "timeout_connect",
    default=TIMEOUT_CONNECT,
    help="Seconds to wait for connections to API",
    type=click.INT,
    show_envvar=True,
    show_default=True,
)
@click.option(
    "--timeout-response",
    "-tr",
    "timeout_response",
    default=TIMEOUT_RESPONSE,
    help="Seconds to wait for responses from API",
    type=click.INT,
    show_default=True,
)
@click.version_option(version.__version__)
@context.pass_context
@click.pass_context
def cli(
    click_ctx,
    ctx,
    log_level_package,
    log_level_http,
    log_level_auth,
    log_level_api,
    log_level_console,
    log_level_file,
    log_console,
    log_file,
    log_request_attrs,
    log_response_attrs,
    log_request_body,
    log_response_body,
    log_file_name,
    log_file_path,
    log_file_max_mb,
    log_file_max_files,
    cert_client_cert,
    cert_client_key,
    cert_client_both,
    proxy,
    certpath,
    certverify,
    certwarn,
    wraperror,
    timeout_connect,
    timeout_response,
    quiet,
):
    """Command line interface for the Axonius API Client."""
    ctx._click_ctx = click_ctx
    ctx.QUIET = quiet
    ctx._connect_args["log_level_package"] = log_level_package
    ctx._connect_args["log_level_http"] = log_level_http
    ctx._connect_args["log_level_auth"] = log_level_auth
    ctx._connect_args["log_level_api"] = log_level_api
    ctx._connect_args["log_level_console"] = log_level_console
    ctx._connect_args["log_level_file"] = log_level_file
    ctx._connect_args["log_console"] = log_console
    ctx._connect_args["log_file"] = log_file
    ctx._connect_args["log_request_attrs"] = log_request_attrs
    ctx._connect_args["log_response_attrs"] = log_response_attrs
    ctx._connect_args["log_request_body"] = log_request_body
    ctx._connect_args["log_response_body"] = log_response_body
    ctx._connect_args["log_file_name"] = log_file_name
    ctx._connect_args["log_file_path"] = log_file_path
    ctx._connect_args["log_file_max_mb"] = log_file_max_mb
    ctx._connect_args["log_file_max_files"] = log_file_max_files
    ctx._connect_args["proxy"] = proxy
    ctx._connect_args["cert_client_cert"] = cert_client_cert
    ctx._connect_args["cert_client_key"] = cert_client_key
    ctx._connect_args["cert_client_both"] = cert_client_both
    ctx._connect_args["certpath"] = certpath
    ctx._connect_args["certverify"] = certverify
    ctx._connect_args["certwarn"] = certwarn
    ctx._connect_args["wraperror"] = wraperror
    ctx._connect_args["timeout_connect"] = timeout_connect
    ctx._connect_args["timeout_response"] = timeout_response


cli.add_command(grp_adapters.adapters)
cli.add_command(grp_assets.devices)
cli.add_command(grp_assets.users)
cli.add_command(grp_system.system)
cli.add_command(grp_tools.tools)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
import click

from ... import context
from . import (cmd_add, cmd_add_from_json, cmd_delete_by_id, cmd_get,
               cmd_get_by_id, cmd_test, cmd_test_by_id)


@click.group(cls=context.AliasedGroup)
def cnx():
    """Group: Work with adapter connections."""


cnx.add_command(cmd_add.cmd)
cnx.add_command(cmd_add_from_json.cmd)
cnx.add_command(cmd_get.cmd)
cnx.add_command(cmd_get_by_id.cmd)
cnx.add_command(cmd_delete_by_id.cmd)
cnx.add_command(cmd_test.cmd)
cnx.add_command(cmd_test_by_id.cmd)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
import click

from ..context import AliasedGroup
from . import (cmd_count, cmd_count_by_saved_query, cmd_destroy, cmd_get,
               cmd_get_by_id, cmd_get_by_saved_query, cmd_get_fields,
               cmd_get_fields_default, cmd_get_tags, grp_saved_query)
from .grp_common import (GET_BY_VALUE_BUILDERS, GET_BY_VALUE_FIELD,
                         GET_BY_VALUE_REGEX_BUILDERS, GET_BY_VALUES_BUILDERS,
                         gen_get_by_cmd)


@click.group(cls=AliasedGroup)
def devices():
    """Group: Work with device assets."""


@click.group(cls=AliasedGroup)
def users():
    """Group: Work with user assets."""


CMDS = [
    grp_saved_query.saved_query,
    cmd_count.cmd,
    cmd_count_by_saved_query.cmd,
    cmd_get_fields.cmd,
    cmd_get_fields_default.cmd,
    cmd_get.cmd,
    cmd_get_by_saved_query.cmd,
    cmd_get_tags.cmd,
    cmd_get_by_id.cmd,
    cmd_destroy.cmd,
]

for cmd in CMDS:
    users.add_command(cmd)
    devices.add_command(cmd)


def add_cmds(grp_obj, fields):
    """Pass."""
    for field in fields:
        grp_obj.add_command(
            gen_get_by_cmd(
                options=GET_BY_VALUE_BUILDERS,
                doc=f"Get assets where {field} equals value",
                cmd_name=f"get-by-{field}",
                method=f"get_by_{field}",
            )
        )

        grp_obj.add_command(
            gen_get_by_cmd(
                options=GET_BY_VALUES_BUILDERS,
                doc=f"Get assets where {field} equals multiple values",
                cmd_name=f"get-by-{field}s",
                method=f"get_by_{field}s",
            )
        )

        grp_obj.add_command(
            gen_get_by_cmd(
                options=GET_BY_VALUE_REGEX_BUILDERS,
                doc=f"Get assets where {field} matches regex value",
                cmd_name=f"get-by-{field}-regex",
                method=f"get_by_{field}_regex",
            )
        )

    grp_obj.add_command(
        gen_get_by_cmd(
            options=[*GET_BY_VALUE_REGEX_BUILDERS, GET_BY_VALUE_FIELD],
            doc=f"Get assets where a field equals value",
            cmd_name=f"get-by-value",
            method=f"get_by_value",
        )
    )

    grp_obj.add_command(
        gen_get_by_cmd(
            options=[*GET_BY_VALUE_REGEX_BUILDERS, GET_BY_VALUE_FIELD],
            doc=f"Get assets where a field equals multiple values",
            cmd_name=f"get-by-values",
            method=f"get_by_values",
        )
    )

    grp_obj.add_command(
        gen_get_by_cmd(
            options=[*GET_BY_VALUE_REGEX_BUILDERS, GET_BY_VALUE_FIELD],
            doc=f"Get assets where a field matches regex value",
            cmd_name=f"get-by-value-regex",
            method=f"get_by_value_regex",
        )
    )


add_cmds(grp_obj=devices, fields=["hostname", "mac", "ip"])
add_cmds(grp_obj=users, fields=["username", "mail"])

devices.add_command(
    gen_get_by_cmd(
        options=GET_BY_VALUE_BUILDERS,
        doc=f"Get assets in subnet",
        cmd_name=f"get-by-subnet",
        method=f"get_by_subnet",
    )
)



# -*- coding: utf-8 -*-
"""API models for working with Enforcement Center."""
from ..mixins import ModelMixins
from ..routers import API_VERSION


class RunAction(ModelMixins):
    """Child API model for working with actions.

    Notes:
        The REST API will need to be updated to allow more power in this library.
        Until then, this class should be considered **BETA**.
    """

    @property
    def router(self):
        """Router for this API model.

        Returns:
            :obj:`.routers.Router`: REST API route defs
        """
        return API_VERSION.actions

    # sort of pointless
    def _get(self):
        """Direct API method to get all actions.

        Returns:
            :obj:`list` of :obj:`str`: all actions known to system
        """
        path = self.router.root
        return self.request(method="get", path=path)

    def _deploy(self, action_name, ids, file_uuid, file_name, params=None):
        """Deploy an action.

        Args:
            name (:obj:`str`): name of action to deploy
            ids (:obj:`list` of :obj:`str`): internal_axon_ids of devices to process
            uuid (:obj:`str`): UUID of binary to use in deployment
            filename (:obj:`str`): filename of binary to use in deployment
            params (:obj:`str`, optional): parameters to pass to action

        Returns:
            :obj:`object`: response from deploying action
        """
        data = {}
        data["action_name"] = action_name
        data["internal_axon_ids"] = ids
        data["binary"] = {}
        data["binary"]["filename"] = file_name
        data["binary"]["uuid"] = file_uuid
        data["params"] = params

        path = self.router.deploy

        return self.request(method="post", path=path, json=data)

    def _shell(self, action_name, ids, command):
        """Run an action.

        Args:
            action_name (:obj:`str`): name of action to run
            ids (:obj:`list` of :obj:`str`): internal_axon_ids of devices to process
            command (:obj:`str`): command to run

        Returns:
            :obj:`object`: response from running action
        """
        data = {}
        data["action_name"] = action_name
        data["internal_axon_ids"] = ids
        data["command"] = command

        path = self.router.shell
        return self.request(method="post", path=path, json=data)

    def _upload_file(self, name, content, content_type=None, headers=None):
        """Upload a file to the system for use in deployment.

        Args:
            binary (:obj:`io.BytesIO`): binary bits of file to upload
            filename (:obj:`str`): name of file to upload

        Returns:
            :obj:`str`: UUID of uploaded file.
        """
        data = {"field_name": "binary"}
        files = {"userfile": (name, content, content_type, headers)}

        path = self.router.upload_file

        ret = self.request(method="post", path=path, data=data, files=files)
        ret["filename"] = name
        return ret



# -*- coding: utf-8 -*-
"""Test suite for axonius_api_client.tools."""
from ....cli import cli
from ....constants import CSV_ADAPTER, DEFAULT_NODE
from ....tools import json_load
from ...utils import load_clirunner


class TestGrpAdaptersFileUpload:
    """Pass."""

    def test_upload(self, request, monkeypatch):
        """Pass."""
        runner = load_clirunner(request, monkeypatch)

        args1 = [
            "adapters",
            "file-upload",
            "--name",
            CSV_ADAPTER,
            "--node-name",
            DEFAULT_NODE,
            "--input-file",
            "-",
        ]
        result1 = runner.invoke(cli=cli, args=args1, input="badwolf")

        stderr1 = result1.stderr
        stdout1 = result1.stdout
        exit_code1 = result1.exit_code

        assert stdout1
        assert stderr1
        assert exit_code1 == 0

        json1 = json_load(stdout1)
        assert isinstance(json1, dict)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ....exceptions import CnxUpdateError
from ...context import CONTEXT_SETTINGS, click
from ...options import AUTH, NODE_CNX, SPLIT_CONFIG_OPT, add_options
from .grp_common import EXPORT, ID_CNX, handle_export

OPTIONS = [
    *AUTH,
    EXPORT,
    *NODE_CNX,
    SPLIT_CONFIG_OPT,
    ID_CNX,
]


@click.command(name="update-by-id", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(
    ctx, url, key, secret, config, adapter_node, adapter_name, cnx_id, **kwargs,
):
    """Update a connection from arguments."""
    client = ctx.obj.start_client(url=url, key=key, secret=secret)
    new_config = dict(config)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        try:
            cnx_new = client.adapters.cnx.update_by_id(
                adapter_name=adapter_name,
                adapter_node=adapter_node,
                cnx_id=cnx_id,
                **new_config,
            )
            ctx.obj.echo_ok(msg=f"Connection updated successfully!")

        except CnxUpdateError as exc:
            ctx.obj.echo_error(msg=f"{exc}", abort=False)
            cnx_new = exc.cnx_new

    handle_export(ctx=ctx, rows=cnx_new, **kwargs)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ..context import CONTEXT_SETTINGS, click
from ..options import (AUTH, EXPORT, FIELDS_SELECT, PAGING, SQ_NAME,
                       add_options, get_option_help)
from .grp_common import GET_EXPORT

METHOD = "get-by-saved-query"
OPTIONS = [
    *AUTH,
    *PAGING,
    *EXPORT,
    *GET_EXPORT,
    *FIELDS_SELECT,
    SQ_NAME,
    get_option_help(choices=["auth", "assetexport", "selectfields"]),
]


@click.command(name="get-by-saved-query", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, **kwargs):
    """Get assets using a saved query."""
    client = ctx.obj.start_client(url=url, key=key, secret=secret)

    p_grp = ctx.parent.command.name
    apiobj = getattr(client, p_grp)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        apiobj.get_by_saved_query(**kwargs)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ....constants import DEFAULT_PERM, PERM_SETS, VALID_PERMS
from ....tools import json_dump, listify
from ...context import click

EXPORT = click.option(
    "--export-format",
    "-xf",
    "export_format",
    type=click.Choice(["json", "str"]),
    help="Format of to export data in",
    default="str",
    show_envvar=True,
    show_default=True,
)

ROLE_NAME = click.option(
    "--name",
    "-n",
    "name",
    help="Name of role",
    required=True,
    show_envvar=True,
    show_default=True,
)

PERMS = [
    click.option(
        f"--{x.lower()}",
        f"-{x.lower()[:4]}",
        f"{x.lower()}",
        help=f"Permission level for {x}",
        type=click.Choice(VALID_PERMS),
        required=False,
        default=DEFAULT_PERM,
        show_envvar=True,
        show_default=True,
    )
    for x in PERM_SETS
]


def handle_export(ctx, data, export_format, **kwargs):
    """Get roles."""
    if export_format == "json":
        click.secho(json_dump(data))
        ctx.exit(0)

    if export_format == "str":
        for item in listify(data):
            name = item["name"]
            perms = item["permissions"]

            click.secho("\n---------------------------------------")
            click.secho(f"Role Name: {name}")
            for k, v in perms.items():
                click.secho(f"{name} -- {k} permission: {v}")
        ctx.exit(0)

    ctx.exit(1)



# -*- coding: utf-8 -*-
"""REST API route definitions."""
from ..tools import join_url


class Router:
    """Simple object store for REST API routes."""

    def __init__(self, object_type, base, version, **routes):
        """Object store for REST API routes.

        Args:
            object_type (:obj:`str`): object type for this set of routes
            base (:obj:`str`): base path for this set of routes
            version (:obj:`int`): api version for this set of routes
            **routes: routes for this object_type
        """
        self._version = version
        self._base = base
        self._object_type = object_type
        self.root = join_url(base, object_type)
        self._routes = ["root"]
        for k, v in routes.items():
            self._routes.append(k)
            setattr(self, k, join_url(self.root, v))

    def __str__(self):
        """Show object info.

        Returns:
            :obj:`str`

        """
        msg = "{obj.__class__.__module__}.{obj.__class__.__name__}"
        msg += "(object_type={obj._object_type!r}, version={obj._version})"
        return msg.format(obj=self)

    def __repr__(self):
        """Show object info.

        Returns:
            :obj:`str`

        """
        return self.__str__()


class ApiV1:
    """Routes provided by the Axonius REST API version 1."""

    version = 1
    base = "api/V{version}".format(version=version)

    users = Router(
        object_type="users",
        base=base,
        version=version,
        by_id="{id}",
        cached="cached",
        count="count",
        views="views",
        view_by_id="views/saved/{id}",
        labels="labels",
        fields="fields",
        destroy="destroy",
    )

    devices = Router(
        object_type="devices",
        base=base,
        cached="cached",
        version=version,
        by_id="{id}",
        count="count",
        views="views",
        view_by_id="views/saved/{id}",
        labels="labels",
        fields="fields",
        destroy="destroy",
    )

    actions = Router(
        object_type="actions",
        base=base,
        version=version,
        shell="shell",  # nosec
        deploy="deploy",
        upload_file="upload_file",
    )

    adapters = Router(
        object_type="adapters",
        base=base,
        version=version,
        cnxs="{adapter_name_raw}/clients",
        cnxs_uuid="{adapter_name_raw}/clients/{cnx_uuid}",
        file_upload="{adapter_name_raw}/{adapter_node_id}/upload_file",
        # file_download="{name}/{node_id}/download_file",
        config_set="{adapter_name_raw}/config/{adapter_config_name}",
        config_get="{adapter_name_plugin}/config/{adapter_config_name}",
    )

    alerts = Router(object_type="alerts", base=base, version=version)

    system = Router(
        object_type="system",
        base=base,
        version=version,
        instances="instances",
        meta_about="meta/about",
        meta_historical_sizes="meta/historical_sizes",
        settings_lifecycle="settings/lifecycle",
        settings_gui="settings/gui",
        settings_core="settings/core",
        discover_lifecycle="discover/lifecycle",
        discover_start="discover/start",
        discover_stop="discover/stop",
        roles_default="roles/default",
        roles="roles",
        roles_by_uuid="roles/{uuid}",
        roles_labels="roles/labels",
        users="users",
        user="users/{uuid}",
        central_core="central_core",
        central_core_restore="central_core/restore",
    )

    all_objects = [
        users,
        devices,
        actions,
        adapters,
        alerts,
        system,
    ]


API_VERSION = ApiV1



# -*- coding: utf-8 -*-
"""API models for working with Enforcement Center."""
from ...tools import listify
from ..mixins import ModelMixins, PagingMixinsObject
from ..routers import API_VERSION


class Enforcements(ModelMixins, PagingMixinsObject):
    """API model for working with enforcement center.

    Notes:
        The REST API will need to be updated to allow more power in this library.
        Until then, this class should be considered **BETA**.
    """

    def delete(self, rows):
        """Delete an enforcement by name."""
        return self._delete(ids=[x["uuid"] for x in listify(obj=rows, dictkeys=False)])

    @property
    def router(self):
        """Router for this API model.

        Returns:
            :obj:`.routers.Router`: REST API route defs
        """
        return API_VERSION.alerts

    def _delete(self, ids):
        """Delete objects by internal axonius IDs.

        Args:
            ids (:obj:`list` of :obj:`str`): internal_axon_ids of devices to process
        """
        path = self.router.root

        return self.request(method="delete", path=path, json=ids)

    def _create(self, name, main, success=None, failure=None, post=None, triggers=None):
        """Create an enforcement set.

        Args:
            name (:obj:`str`): name of enforcement to create
            main (:obj:`dict`): main action
            success (:obj:`list` of :obj:`dict`, optional): success actions
            failure (:obj:`list` of :obj:`dict`, optional): failure actions
            post (:obj:`list` of :obj:`dict`, optional): post actions
            triggers (:obj:`list` of :obj:`dict`, optional): saved query trigger

        Notes:
            This will get a public create method once the REST API server has been
            updated to expose /enforcements/actions, /api/enforcements/actions/saved,
            and others.

        Returns:
            :obj:`str`: ID of created enforcement set.
        """
        data = {}
        data["name"] = name
        data["actions"] = {}
        data["actions"]["main"] = main
        data["actions"]["success"] = success or []
        data["actions"]["failure"] = failure or []
        data["actions"]["post"] = post or []
        data["triggers"] = triggers or []

        path = self.router.root
        return self.request(method="put", path=path, json=data, is_json=False)

    def _get(self, query=None, row_start=0, page_size=0):
        """Get a page for a given query."""
        params = {}
        params["skip"] = row_start
        params["limit"] = page_size
        params["filter"] = query

        path = self.router.root

        return self.request(method="get", path=path, params=params)



# -*- coding: utf-8 -*-
"""Test suite for axonapi.api.enforcements."""
import pytest

from axonius_api_client.exceptions import NotFoundError

from ...meta import CREATE_EC_ACTION_MAIN, CREATE_EC_NAME, CREATE_EC_TRIGGER1


class TestEnforcementsBase:
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_enforcements):
        """Pass."""
        return api_enforcements


class TestEnforcementsPrivate(TestEnforcementsBase):
    """Pass."""

    def test_private_get(self, apiobj):
        """Pass."""
        data = apiobj._get()
        assert isinstance(data, dict)

        assets = data["assets"]
        assert isinstance(assets, list)

        for asset in assets:
            assert isinstance(asset, dict)


class TestEnforcementsPublic(TestEnforcementsBase):
    """Pass."""

    def test_get(self, apiobj):
        """Pass."""
        data = apiobj.get()
        assert isinstance(data, list)
        for found in data:
            assert isinstance(found["uuid"], str)
            assert isinstance(found["actions.main"], str)
            assert isinstance(found["name"], str)
            assert isinstance(found["date_fetched"], str)
            assert isinstance(found["last_updated"], str)
            assert "triggers.last_triggered" in found
            assert "triggers.times_triggered" in found

    def test_get_maxpages(self, apiobj):
        """Pass."""
        found = apiobj.get(max_pages=1, page_size=1)
        assert isinstance(found, list)
        # we can't test for length if there are no enforcements...
        # assert len(found) == 1

    def test_create_get_delete(self, apiobj, api_users):
        """Pass."""
        try:
            old_found = apiobj.get_by_name(CREATE_EC_NAME, eq_single=False)
        except Exception:
            old_found = None
        else:
            deleted = apiobj.delete(rows=old_found)
            assert isinstance(deleted, dict)
            assert isinstance(deleted["deleted"], int)
            assert deleted["deleted"] == 1

        trigger_name = api_users.saved_query.get()[0]["name"]
        trigger = {"view": {"name": trigger_name, "entity": "users"}}
        trigger.update(CREATE_EC_TRIGGER1)

        created = apiobj._create(
            name=CREATE_EC_NAME, main=CREATE_EC_ACTION_MAIN, triggers=[trigger],
        )
        assert isinstance(created, str)

        found = apiobj.get_by_name(CREATE_EC_NAME)
        """
        {
            "actions.main": "Badwolf Create Notification",
            "date_fetched": "2019-09-10 23:17:07+00:00",
            "last_updated": "Tue, 10 Sep 2019 23:17:07 GMT",
            "name": "Badwolf EC Example",
            "triggers.last_triggered": null,
            "triggers.times_triggered": 0,
            "triggers.view.name": "Users Created in Last 30 Days",
            "uuid": "5d782ef380ded0001bbe3c47"
        }
        """
        assert isinstance(found, dict)
        assert found["uuid"] == created
        assert found["actions.main"] == CREATE_EC_ACTION_MAIN["name"]
        assert found["name"] == CREATE_EC_NAME
        assert isinstance(found["date_fetched"], str)
        assert isinstance(found["last_updated"], str)
        assert "triggers.last_triggered" in found
        assert "triggers.times_triggered" in found
        assert found["triggers.view.name"] == trigger_name

        found_by_id = apiobj.get_by_uuid(found["uuid"])
        assert isinstance(found_by_id, dict)

        deleted = apiobj.delete(rows=found_by_id)
        assert isinstance(deleted, dict)
        assert isinstance(deleted["deleted"], int)
        assert deleted["deleted"] == 1

        with pytest.raises(NotFoundError):
            apiobj.get_by_uuid(found["uuid"])

        with pytest.raises(NotFoundError):
            apiobj.get_by_name(found["name"])

        allobjs = apiobj.get()
        if allobjs:
            name = allobjs[0]["name"]
            rematches = apiobj.get_by_name(value=name[0], value_regex=True)
            assert any([x["name"] == name for x in rematches])



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
import click

from ..constants import DEFAULT_NODE, DEFAULT_PATH, MAX_PAGE_SIZE
from ..tools import coerce_int
from . import context
from .helps import HELPSTRS


def add_options(options):
    """Pass."""

    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


def int_callback(ctx, param, value):
    """Pass."""
    return coerce_int(value)


def help_callback(ctx, param, value):
    """Pass."""
    if value:
        helpstr = HELPSTRS[value]
        click.secho(helpstr, err=True, fg="blue")
        ctx.exit(0)


def get_option_help(choices):
    """Pass."""
    return click.option(
        "--help-detailed",
        "help_detailed",
        help="Show detailed help and exit",
        type=click.Choice(choices),
        callback=help_callback,
        is_eager=True,
    )


def get_option_fields_default(default=True):
    """Pass."""
    return click.option(
        "--fields-default/--no-fields-default",
        "-fd/-nfd",
        "fields_default",
        default=default,
        help="Include the default fields defined in the API library",
        is_flag=True,
        show_envvar=True,
        show_default=True,
    )


URL = click.option(
    "--url",
    "-u",
    "url",
    required=True,
    help="URL of an Axonius instance",
    metavar="URL",
    prompt="URL",
    show_envvar=True,
    show_default=True,
)


AUTH = [
    URL,
    click.option(
        "--key",
        "-k",
        "key",
        required=True,
        help="API Key of user in an Axonius instance",
        metavar="KEY",
        prompt="API Key of user",
        hide_input=True,
        show_envvar=True,
        show_default=True,
    ),
    click.option(
        "--secret",
        "-s",
        "secret",
        required=True,
        help="API Secret of user in an Axonius instance",
        metavar="SECRET",
        prompt="API Secret of user",
        hide_input=True,
        show_envvar=True,
        show_default=True,
    ),
]

SQ_NAME = click.option(
    "--name",
    "-n",
    "name",
    help="Name of saved query",
    required=True,
    show_envvar=True,
    show_default=True,
)

QUERY = [
    click.option(
        "--query",
        "-q",
        "query",
        help="Query built from the Query wizard in the GUI",
        metavar="QUERY",
        show_envvar=True,
        show_default=True,
    ),
    click.option(
        "--query-file",
        "-qf",
        "query_file",
        help="Path to a file to override --query",
        type=click.File(),
        metavar="QUERY_FILE",
        show_envvar=True,
        show_default=True,
    ),
]

FIELDS_SELECT = [
    click.option(
        "--field",
        "-f",
        "fields",
        help="Fields to include in the format of adapter:field (multiples)",
        metavar="ADAPTER:FIELD",
        multiple=True,
        show_envvar=True,
        show_default=True,
    ),
    click.option(
        "--field-regex",
        "-fr",
        "fields_regex",
        help="Regular expressions of fields to include (multiples)",
        metavar="ADAPTER_REGEX:FIELD_REGEX",
        multiple=True,
        show_envvar=True,
        show_default=True,
    ),
]

EXPORT = [
    click.option(
        "--export-file",
        "-xf",
        "export_file",
        default="",
        help="File to send data to",
        show_envvar=True,
        show_default=True,
        metavar="PATH",
    ),
    click.option(
        "--export-path",
        "-xp",
        "export_path",
        default=DEFAULT_PATH,
        help="If --export-file supplied, the directory to write --export_file to",
        type=click.Path(exists=False, resolve_path=True),
        show_envvar=True,
        show_default=True,
    ),
    click.option(
        "--export-overwrite/--no-export-overwrite",
        "-xo/-nxo",
        "export_overwrite",
        default=False,
        help="If --export-file supplied and it exists, overwrite it",
        is_flag=True,
        show_envvar=True,
        show_default=True,
    ),
]

PAGING = [
    click.option(
        "--max-rows",
        "max_rows",
        help="Stop fetching assets after this many rows have been received",
        type=click.INT,
        show_envvar=True,
        show_default=True,
    ),
    click.option(
        "--max-pages",
        "max_pages",
        help="Stop fetching assets after this many pages have been received",
        type=click.INT,
        show_envvar=True,
        show_default=True,
    ),
    click.option(
        "--page-size",
        "page_size",
        help="Number of assets to fetch per page",
        default=MAX_PAGE_SIZE,
        type=click.INT,
        show_envvar=True,
        show_default=True,
    ),
    click.option(
        "--page-start",
        "page_start",
        help="Start fetching from page N of --page-size",
        default=0,
        type=click.INT,
        show_envvar=True,
        show_default=True,
    ),
    click.option(
        "--page-sleep",
        "page_sleep",
        default=0,
        type=click.INT,
        help="Seconds to wait in between each fetch of a page",
        show_envvar=True,
        show_default=True,
    ),
]

SPLIT_CONFIG_OPT = click.option(
    "--config",
    "-c",
    "config",
    help="Configuration keys in the form of key=value (multiples)",
    type=context.SplitEquals(),
    multiple=True,
    show_envvar=True,
    show_default=True,
    required=False,
)

SPLIT_CONFIG_REQ = click.option(
    "--config",
    "-c",
    "config",
    help="Configuration keys in the form of key=value (multiples)",
    type=context.SplitEquals(),
    multiple=True,
    show_envvar=True,
    show_default=True,
    required=True,
)

INPUT_FILE = click.option(
    "--input-file",
    "-if",
    "input_file",
    help="File to read (from path or piped via STDIN)",
    default="-",
    type=click.File(mode="r"),
    show_envvar=True,
    show_default=True,
)
NODE = [
    click.option(
        "--node-name",
        "-nn",
        "node",
        default=DEFAULT_NODE,
        show_envvar=True,
        show_default=True,
        help="Node name",
    ),
    click.option(
        "--name",
        "-n",
        "name",
        required=True,
        show_envvar=True,
        show_default=True,
        help="Adapter name",
    ),
]


NODE_CNX = [
    click.option(
        "--node-name",
        "-nn",
        "adapter_node",
        default=DEFAULT_NODE,
        show_envvar=True,
        show_default=True,
        help="Node name",
    ),
    click.option(
        "--name",
        "-n",
        "adapter_name",
        required=True,
        show_envvar=True,
        show_default=True,
        help="Adapter name",
    ),
]



# -*- coding: utf-8 -*-
"""API model for working with system configuration."""
from ...exceptions import ApiError, NotFoundError
from ..mixins import ChildMixins
from ..parsers.config import parse_unchanged


class Users(ChildMixins):
    """User Role controls."""

    def add(
        self, name, password, role_name, first_name=None, last_name=None, email=None,
    ):
        """Pass."""
        role = self.parent.roles.get_by_name(name=role_name)
        users = self.get()

        names = [x["user_name"] for x in users]

        if name in names:
            raise ApiError(f"User named {name!r} already exists")

        user = {}
        user["user_name"] = name
        user["first_name"] = first_name
        user["last_name"] = last_name
        user["password"] = password
        user["email"] = email
        user["role_id"] = role["uuid"]
        user["role_obj"] = role
        self._add(user=user)
        return self.get_by_name(name=name)

    def get(self):
        """Pass."""
        users = self._get()
        roles = self.parent.roles.get()
        for user in users:
            for role in roles:
                if role["uuid"] == user["role_id"]:
                    user["role_obj"] = role
                    break

        return users

    def get_by_name(self, name):
        """Pass."""
        users = self.get()

        valid = [x["user_name"] for x in users]

        if name not in valid:
            valid = "\n" + "\n".join(valid)
            raise NotFoundError(f"User name {name!r} not found, valid users:{valid}")

        return [x for x in users if x["user_name"] == name][0]

    def update(
        self,
        name,
        first_name=None,
        last_name=None,
        password=None,
        email=None,
        role_name=None,
    ):
        """Pass."""
        user = self.get_by_name(name=name)

        one_of = [first_name, last_name, password, email, role_name]

        if all([x is None for x in one_of]):
            req = ", ".join(
                ["first_name", "last_name", "password", "email", "role_name"]
            )
            raise ApiError(f"Must supply at least one of: {req!r}")

        uuid = user["uuid"]

        if first_name is not None:
            user["first_name"] = first_name

        if last_name is not None:
            user["last_name"] = last_name

        if password is not None:
            _, password = parse_unchanged(value=password)
            user["password"] = password

        if email is not None:
            user["email"] = email

        if role_name is not None:
            role = self.parent.roles.get_by_name(name=role_name)
            user["role_id"] = role["uuid"]
            user["role_obj"] = role

        self._update(uuid=uuid, user=user)
        return self.get_by_name(name=name)

    def delete(self, name):
        """Pass."""
        user = self.get_by_name(name=name)
        return self._delete(uuid=user["uuid"])

    def _get(self, limit=None, skip=None):
        """Pass."""
        data = {}

        if limit is not None:
            data["limit"] = limit

        if skip is not None:
            data["skip"] = skip

        path = self.router.users
        return self.request(method="get", path=path, params=data)

    def _add(self, user):
        """Pass."""
        path = self.router.users
        return self.request(method="put", path=path, json=user)

    def _delete(self, uuid):
        """Pass."""
        path = self.router.user.format(uuid=uuid)
        return self.request(method="delete", path=path)

    def _update(self, uuid, user):
        """Pass."""
        path = self.router.user.format(uuid=uuid)
        return self.request(
            method="post", path=path, json=user, error_json_invalid=False
        )



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ...context import CONTEXT_SETTINGS, click
from ...options import AUTH, add_options
from .grp_common import EXPORT, handle_export

USER_NAME = click.option(
    "--name",
    "-n",
    "name",
    help="Name of user",
    required=True,
    show_envvar=True,
    show_default=True,
)
PASSWORD = click.option(
    "--password",
    "-p",
    "password",
    help="Name of user",
    required=True,
    show_envvar=True,
    show_default=True,
)
ROLE_NAME = click.option(
    "--role-name",
    "-r",
    "role_name",
    help="Role to assign to user",
    required=True,
    show_envvar=True,
    show_default=True,
)
FIRST_NAME = click.option(
    "--first-name",
    "-f",
    "first_name",
    help="First name of user",
    required=False,
    show_envvar=True,
    show_default=True,
)
LAST_NAME = click.option(
    "--last-name",
    "-l",
    "last_name",
    help="Last name of user",
    required=False,
    show_envvar=True,
    show_default=True,
)
EMAIL = click.option(
    "--email",
    "-e",
    "email",
    help="Email address of user",
    required=False,
    show_envvar=True,
    show_default=True,
)

OPTIONS = [*AUTH, EXPORT, FIRST_NAME, LAST_NAME, EMAIL, USER_NAME, PASSWORD, ROLE_NAME]


@click.command(name="add", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, export_format, name, **kwargs):
    """Add a user."""
    client = ctx.obj.start_client(url=url, key=key, secret=secret)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        data = client.system.users.add(name=name, **kwargs)
        ctx.obj.echo_ok(f"Added user {name!r}")

    handle_export(ctx=ctx, data=data, export_format=export_format, **kwargs)



# -*- coding: utf-8 -*-
"""Test suite for assets."""
import pytest

from axonius_api_client.constants import AGG_ADAPTER_ALTS, AGG_ADAPTER_NAME
from axonius_api_client.exceptions import ApiError, NotFoundError

from ...meta import FIELD_FORMATS, NORM_TYPES, SCHEMA_FIELD_FORMATS, SCHEMA_TYPES


def load_test_data(apiobj):
    """Pass."""
    apiobj.TEST_DATA = getattr(apiobj, "TEST_DATA", {})

    if not apiobj.TEST_DATA.get("fields_map"):
        apiobj.TEST_DATA["fields_map"] = apiobj.fields.get()
    return apiobj


class FieldsPrivate:
    """Pass."""

    def test_private_get(self, apiobj):
        """Pass."""
        fields = apiobj.fields._get()
        self.val_raw_fields(fields=fields)
        assert not fields

    def val_raw_fields(self, fields):
        """Pass."""
        assert isinstance(fields, dict)

        schema = fields.pop("schema")
        assert isinstance(schema, dict)

        generic = fields.pop("generic")
        assert isinstance(generic, list)
        self.val_raw_adapter_fields(adapter="generic", adapter_fields=generic)

        generic_schema = schema.pop("generic")
        assert isinstance(generic_schema, dict)
        self.val_raw_schema(adapter="generic", schema=generic_schema)

        specific = fields.pop("specific")
        assert isinstance(specific, dict)

        specific_schema = schema.pop("specific")
        assert isinstance(specific_schema, dict)

        for adapter, adapter_fields in specific.items():
            self.val_raw_adapter_fields(adapter=adapter, adapter_fields=adapter_fields)
            adapter_schema = specific_schema.pop(adapter)
            self.val_raw_schema(adapter=adapter, schema=adapter_schema)

        assert not fields
        assert not schema

    def val_raw_schema(self, adapter, schema):
        """Pass."""
        assert isinstance(schema, dict)

        items = schema.pop("items")
        assert isinstance(items, list)

        required = schema.pop("required")
        assert isinstance(required, list)

        stype = schema.pop("type")
        assert stype == "array"

        assert not schema, list(schema)

        for req in required:
            assert isinstance(req, str)

        for item in items:
            assert item
            self.val_raw_items(adapter=adapter, items=item)

    def val_raw_adapter_fields(self, adapter, adapter_fields):
        """Pass."""
        assert isinstance(adapter_fields, list)

        for field in adapter_fields:
            assert isinstance(field, dict)

            name = field.pop("name")
            assert isinstance(name, str) and name

            title = field.pop("title")
            assert isinstance(title, str) and title

            ftype = field.pop("type")
            assert isinstance(ftype, str) and ftype
            assert ftype in SCHEMA_TYPES

            description = field.pop("description", "")
            assert isinstance(description, str)

            sort = field.pop("sort", False)
            assert isinstance(sort, bool)

            unique = field.pop("unique", False)
            assert isinstance(unique, bool)

            branched = field.pop("branched", False)
            assert isinstance(branched, bool)

            dynamic = field.pop("dynamic", False)
            assert isinstance(dynamic, bool)

            fformat = field.pop("format", "")
            assert isinstance(fformat, str)
            assert fformat in FIELD_FORMATS or fformat == ""

            enums = field.pop("enum", [])
            assert isinstance(enums, list)
            for enum in enums:
                assert isinstance(enum, str) or isinstance(enum, int)

            items = field.pop("items", {})
            assert isinstance(items, dict)

            self.val_raw_items(adapter=f"{adapter}:{name}", items=items)

            assert not field, list(field)

    def val_raw_items(self, adapter, items):
        """Pass."""
        assert isinstance(items, dict)

        if items:
            ftype = items.pop("type")
            assert isinstance(ftype, str)
            assert ftype in SCHEMA_TYPES

            description = items.pop("description", "")
            assert isinstance(description, str)

            title = items.pop("title", "")
            assert isinstance(title, str)

            name = items.pop("name", "")
            assert isinstance(name, str)

            sort = items.pop("sort", False)
            assert isinstance(sort, bool)

            unique = items.pop("unique", False)
            assert isinstance(unique, bool)

            branched = items.pop("branched", False)
            assert isinstance(branched, bool)

            dynamic = items.pop("dynamic", False)
            assert isinstance(dynamic, bool)

            hidden = items.pop("hidden", False)
            assert isinstance(hidden, bool)

            iformat = items.pop("format", "")
            assert isinstance(iformat, str)
            assert iformat in SCHEMA_FIELD_FORMATS or iformat == ""

            val_source(obj=items)

            enums = items.pop("enum", [])
            assert isinstance(enums, list)

            for enum in enums:
                assert isinstance(enum, str) or isinstance(enum, int)

            sub_items = items.pop("items", [])
            assert isinstance(sub_items, list) or isinstance(sub_items, dict)
            assert not items, list(items)

            if isinstance(sub_items, dict):
                self.val_raw_items(adapter=adapter, items=sub_items)
            else:
                for sub_item in sub_items:
                    self.val_raw_items(adapter=adapter, items=sub_item)

    def test_private_prettify_schemas(self, apiobj):
        """Pass."""
        fields_map = apiobj.TEST_DATA["fields_map"]
        schemas = fields_map[AGG_ADAPTER_NAME]
        pretty = apiobj.fields._prettify_schemas(schemas=schemas)
        assert isinstance(pretty, list)
        for p in pretty:
            assert isinstance(p, str)
            assert "->" in p


class FieldsPublic:
    """Pass."""

    def test_get(self, apiobj):
        """Pass."""
        fields_map = apiobj.fields.get()
        assert isinstance(fields_map, dict)
        assert isinstance(fields_map[AGG_ADAPTER_NAME], list)
        self.val_parsed_fields(fields_map=fields_map)

    def val_parsed_fields(self, fields_map):
        """Pass."""
        assert isinstance(fields_map, dict)

        for adapter, schemas in fields_map.items():
            assert not adapter.endswith("_adapter")
            assert isinstance(schemas, list)

            for schema in schemas:
                self.val_parsed_schema(schema=schema, adapter=adapter)

    def val_parsed_schema(self, schema, adapter):
        """Pass."""
        assert isinstance(schema, dict)

        name = schema.pop("name")
        assert isinstance(name, str) and name

        ftype = schema.pop("type")
        assert isinstance(ftype, str) and ftype
        assert ftype in SCHEMA_TYPES

        fformat = schema.pop("format", "")
        assert isinstance(fformat, str)
        assert fformat in FIELD_FORMATS or fformat == ""

        adapter_name = schema.pop("adapter_name")
        assert isinstance(adapter_name, str) and adapter_name

        adapter_name_raw = schema.pop("adapter_name_raw")
        assert isinstance(adapter_name_raw, str) and adapter_name_raw

        adapter_prefix = schema.pop("adapter_prefix")
        assert isinstance(adapter_prefix, str) and adapter_prefix

        adapter_title = schema.pop("adapter_title")
        assert isinstance(adapter_title, str) and adapter_title

        column_name = schema.pop("column_name")
        assert isinstance(column_name, str) and column_name

        column_title = schema.pop("column_title")
        assert isinstance(column_title, str) and column_title

        name_base = schema.pop("name_base")
        assert isinstance(name_base, str) and name_base

        name_qual = schema.pop("name_qual")
        assert isinstance(name_qual, str) and name_qual

        title = schema.pop("title")
        assert isinstance(title, str) and title

        type_norm = schema.pop("type_norm")
        assert isinstance(type_norm, str) and type_norm
        assert type_norm in NORM_TYPES

        parent = schema.pop("parent")
        assert isinstance(parent, str) and parent

        is_root = schema.pop("is_root")
        assert isinstance(is_root, bool)

        is_list = schema.pop("is_list")
        assert isinstance(is_list, bool)

        description = schema.pop("description", "")
        assert isinstance(description, str)

        sort = schema.pop("sort", False)
        assert isinstance(sort, bool)

        unique = schema.pop("unique", False)
        assert isinstance(unique, bool)

        branched = schema.pop("branched", False)
        assert isinstance(branched, bool)

        dynamic = schema.pop("dynamic", False)
        assert isinstance(dynamic, bool)

        is_complex = schema.pop("is_complex")
        assert isinstance(is_complex, bool)

        enums = schema.pop("enum", [])
        assert isinstance(enums, list)

        for enum in enums:
            assert isinstance(enum, str) or isinstance(enum, int)

        if is_complex:
            sub_fields = schema.pop("sub_fields", {})
            assert isinstance(sub_fields, list)
            if name != "all":
                assert sub_fields

            for sub_field in sub_fields:
                self.val_parsed_schema(adapter=f"{adapter}:{name}", schema=sub_field)
        else:
            items = schema.pop("items", {})

            dynamic = items.pop("dynamic", False)
            assert isinstance(dynamic, bool)

            iformat = items.pop("format", "")
            assert isinstance(iformat, str)
            assert iformat in SCHEMA_FIELD_FORMATS or iformat == ""

            itype = items.pop("type", "")
            assert isinstance(itype, str)
            assert itype in SCHEMA_TYPES or itype == ""

            val_source(obj=items)

            enums = items.pop("enum", [])
            assert isinstance(enums, list)

            for enum in enums:
                assert isinstance(enum, str) or isinstance(enum, int)

            assert not items

        assert not schema, list(schema)

    def test_get_adapter_name(self, apiobj):
        """Pass."""
        search = AGG_ADAPTER_ALTS[0]
        exp = AGG_ADAPTER_NAME
        adapter = apiobj.fields.get_adapter_name(
            value=search, fields_map=apiobj.TEST_DATA["fields_map"]
        )
        assert adapter == exp

    def test_get_adapter_name_error(self, apiobj):
        """Pass."""
        search = "badwolf"
        with pytest.raises(NotFoundError):
            apiobj.fields.get_adapter_name(
                value=search, fields_map=apiobj.TEST_DATA["fields_map"]
            )

    def test_get_adapter_names_single(self, apiobj):
        """Pass."""
        search = AGG_ADAPTER_ALTS[0]
        exp = [AGG_ADAPTER_NAME]
        adapters = apiobj.fields.get_adapter_names(
            value=search, fields_map=apiobj.TEST_DATA["fields_map"]
        )
        assert adapters == exp

    def test_get_adapter_names_multi(self, apiobj):
        """Pass."""
        search = "a"
        adapters = apiobj.fields.get_adapter_names(
            value=search, fields_map=apiobj.TEST_DATA["fields_map"]
        )
        assert AGG_ADAPTER_NAME in adapters
        assert len(adapters) > 1

    def test_get_adapter_names_error(self, apiobj):
        """Pass."""
        search = "badwolf"
        with pytest.raises(NotFoundError):
            apiobj.fields.get_adapter_names(
                value=search, fields_map=apiobj.TEST_DATA["fields_map"]
            )

    def test_get_field_schema(self, apiobj):
        """Pass."""
        search = "last_seen"
        exp = [
            x
            for x in apiobj.TEST_DATA["fields_map"][AGG_ADAPTER_NAME]
            if x["name_base"] == search
        ][0]
        result = apiobj.fields.get_field_schema(
            value=search, schemas=apiobj.TEST_DATA["fields_map"][AGG_ADAPTER_NAME]
        )
        assert exp == result

    def test_get_field_names_re(self, apiobj):
        """Pass."""
        search = ["seen"]
        result = apiobj.fields.get_field_names_re(
            value=search, fields_map=apiobj.TEST_DATA["fields_map"]
        )
        assert "specific_data.data.last_seen" in result

    def test_get_field_names_eq(self, apiobj):
        """Pass."""
        search = ["specific_data.data.id", "last_seen"]
        exp = []
        for i in search:
            exp += [
                x["name_qual"]
                for x in apiobj.TEST_DATA["fields_map"][AGG_ADAPTER_NAME]
                if x["name_base"] == i or x["name_qual"] == i
            ]
        result = apiobj.fields.get_field_names_eq(
            value=search, fields_map=apiobj.TEST_DATA["fields_map"]
        )
        assert exp == result

    def test_get_field_schemas(self, apiobj):
        """Pass."""
        search = "l"
        exp = [
            x
            for x in apiobj.TEST_DATA["fields_map"][AGG_ADAPTER_NAME]
            if search in x["name_base"]
        ]
        result = apiobj.fields.get_field_schemas(
            value=search, schemas=apiobj.TEST_DATA["fields_map"][AGG_ADAPTER_NAME]
        )
        assert exp == result

    def test_get_field_schema_error(self, apiobj):
        """Pass."""
        search = "badwolf"
        with pytest.raises(NotFoundError):
            apiobj.fields.get_field_schema(
                value=search, schemas=apiobj.TEST_DATA["fields_map"][AGG_ADAPTER_NAME],
            )

    def test_get_field_schemas_error(self, apiobj):
        """Pass."""
        search = "badwolf"
        with pytest.raises(NotFoundError):
            apiobj.fields.get_field_schemas(
                value=search, schemas=apiobj.TEST_DATA["fields_map"][AGG_ADAPTER_NAME],
            )

    @pytest.mark.parametrize(
        "test_data",
        [
            (f"{AGG_ADAPTER_NAME}:host", (AGG_ADAPTER_NAME, ["host"])),
            (
                f"{AGG_ADAPTER_NAME}:host, ip, other",
                (AGG_ADAPTER_NAME, ["host", "ip", "other"]),
            ),
            ("host, ip, other", (AGG_ADAPTER_NAME, ["host", "ip", "other"])),
            ("adapter1:host, ip, other", ("adapter1", ["host", "ip", "other"])),
            (":host", (AGG_ADAPTER_NAME, ["host"])),
        ],
        scope="class",
    )
    def test_split_search(self, apiobj, test_data):
        """Pass."""
        search, exp = test_data
        result = apiobj.fields.split_search(value=search)
        assert result == exp

    @pytest.mark.parametrize(
        "test_data",
        [
            (f"{AGG_ADAPTER_NAME}:host", [(AGG_ADAPTER_NAME, ["host"])]),
            (
                f"{AGG_ADAPTER_NAME}:host, ip, other",
                [(AGG_ADAPTER_NAME, ["host", "ip", "other"])],
            ),
            ("host, ip, other", [(AGG_ADAPTER_NAME, ["host", "ip", "other"])]),
            ("adapter1:host, ip, other", [("adapter1", ["host", "ip", "other"])]),
            (
                [f"{AGG_ADAPTER_NAME}:host", "adapter1:host, ip, other"],
                [(AGG_ADAPTER_NAME, ["host"]), ("adapter1", ["host", "ip", "other"])],
            ),
        ],
        scope="class",
    )
    def test_split_searches(self, apiobj, test_data):
        """Pass."""
        searches, exp = test_data
        result = apiobj.fields.split_searches(value=searches)
        assert result == exp

    def test_split_search_error(self, apiobj):
        """Pass."""
        search = f"{AGG_ADAPTER_NAME}:"
        with pytest.raises(ApiError):
            apiobj.fields.split_search(value=search)

    def test_get_field_name_manual(self, apiobj):
        """Pass."""
        search = "test"
        result = apiobj.fields.get_field_name(
            value=search, fields_map=apiobj.TEST_DATA["fields_map"], field_manual=True,
        )
        assert search == result

    def test_get_field_name_error(self, apiobj):
        """Pass."""
        search = "bad,wolf"
        with pytest.raises(ApiError):
            apiobj.fields.get_field_name(
                value=search, fields_map=apiobj.TEST_DATA["fields_map"]
            )

    def test_get_field_name(self, apiobj):
        """Pass."""
        search = "last_seen"
        exp = "specific_data.data.last_seen"
        result = apiobj.fields.get_field_name(
            value=search, fields_map=apiobj.TEST_DATA["fields_map"]
        )
        assert result == exp

    def test_validate(self, apiobj):
        """Pass."""
        exp = apiobj.fields_default + [
            "specific_data.data",
            "specific_data.data.last_seen",
            "specific_data.data.first_fetch_time",
        ]
        result = apiobj.fields.validate(
            fields=["last_seen"],
            fields_regex=["^first_fetch_time$"],
            fields_manual=["specific_data.data"],
            fields_default=True,
        )
        assert result == exp

    def test_validate_defaults(self, apiobj):
        """Pass."""
        exp = apiobj.fields_default
        result = apiobj.fields.validate(
            fields=None, fields_regex=None, fields_manual=None, fields_default=True
        )
        assert exp == result

    def test_validate_error(self, apiobj):
        """Pass."""
        with pytest.raises(ApiError):
            apiobj.fields.validate(
                fields=None, fields_regex=None, fields_manual=None, fields_default=False,
            )


class TestFieldsDevices(FieldsPrivate, FieldsPublic):
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_devices):
        """Pass."""
        return load_test_data(api_devices)


class TestFieldsUsers(FieldsPrivate, FieldsPublic):
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_users):
        """Pass."""
        return load_test_data(api_users)


def val_source(obj):
    """Pass."""
    source = obj.pop("source", {})
    assert isinstance(source, dict)

    if source:
        source_key = source.pop("key")
        assert isinstance(source_key, str)

        source_options = source.pop("options")
        assert isinstance(source_options, dict)

        options_allow = source_options.pop("allow-custom-option")
        assert isinstance(options_allow, bool)

        assert not source, source
        assert not source_options, source_options



# -*- coding: utf-8 -*-
"""API models for working with adapters and connections."""
import copy

from ...constants import NO, SETTING_UNCHANGED, YES
from ...exceptions import (ApiError, ConfigInvalidValue, ConfigRequired,
                           ConfigUnchanged, ConfigUnknown)
from ...tools import is_int, json_load
from .tables import tablize_schemas


def config_check(value, schema, source, callbacks=None, none_ok=True):
    """Check a supplied value for a setting is correctly typed."""
    schema_type = schema["type"]

    if value is None or str(value).strip() == "None":
        if none_ok:
            return None
        else:
            sinfo = config_info(schema=schema, value=value, source=source)
            raise ConfigInvalidValue(f"{sinfo}\nValue of None not allowed")

    if schema_type == "file":
        return config_check_file(
            value=value, schema=schema, callbacks=callbacks, source=source
        )

    if schema_type == "bool":
        return config_check_bool(
            value=value, schema=schema, callbacks=callbacks, source=source
        )

    if schema_type == "number":
        return config_check_int(
            value=value, schema=schema, callbacks=callbacks, source=source
        )

    if schema_type == "integer":
        return config_check_int(
            value=value, schema=schema, callbacks=callbacks, source=source
        )

    if schema_type == "array":
        return config_check_array(
            value=value, schema=schema, callbacks=callbacks, source=source
        )

    if schema_type == "string":
        return config_check_str(
            value=value, schema=schema, callbacks=callbacks, source=source
        )

    valids = ["string", "integer", "number", "bool", "array"]
    valids = ", ".join(valids)
    raise ApiError(f"Schema type {schema_type!r} is unknown, valids: {valids}")


def config_check_file(value, schema, source, callbacks=None):
    """Pass."""
    sname = schema["name"]

    callbacks = callbacks or {}
    cb_file = callbacks.get("cb_file", None)

    is_file, value = is_uploaded_file(value=value)
    if is_file:
        return value

    if callable(cb_file):
        return cb_file(value=value, schema=schema, callbacks=callbacks, source=source)

    raise ApiError(f"File uploads for {source} setting {sname!r} are not supported yet")


def config_check_bool(value, schema, source, callbacks=None):
    """Pass."""
    coerce = value.lower().strip() if isinstance(value, str) else value

    if coerce in YES:
        return True

    if coerce in NO:
        return False

    sinfo = config_info(schema=schema, value=value, source=source)

    yes = ", ".join(sorted([str(x) for x in YES]))
    no = ", ".join(sorted([str(x) for x in NO]))
    msg = [
        sinfo,
        "Is not a valid boolean!",
        f"Valid values for true: {yes}",
        f"Valid values for false: {no}",
    ]
    raise ConfigInvalidValue("\n".join(msg))


def config_check_int(value, schema, source, callbacks=None):
    """Pass."""
    if is_int(obj=value, digit=True):
        return int(value)

    sinfo = config_info(schema=schema, value=value, source=source)
    raise ConfigInvalidValue(f"{sinfo}\nIs not a valid integer!")


def config_check_array(value, schema, source, callbacks=None):
    """Pass."""
    if isinstance(value, str):
        value = [x.strip() for x in value.split(",") if x.strip()]

    is_list = isinstance(value, list)

    if not is_list or (is_list and not all([isinstance(x, str) for x in value])):
        sinfo = config_info(schema=schema, value=value, source=source)
        msg = f"{sinfo}\nIs not a list of strings or a comma seperated string!"
        raise ConfigInvalidValue(msg)

    return value


def parse_unchanged(value):
    """Pass."""
    unchanges = [
        SETTING_UNCHANGED,
        str(SETTING_UNCHANGED),
        SETTING_UNCHANGED[0],
        str(SETTING_UNCHANGED[0]),
    ]
    if value in unchanges:
        return True, SETTING_UNCHANGED
    return False, value


def config_check_str(value, schema, source, callbacks=None):
    """Pass."""
    schema_fmt = schema.get("format", "")
    schema_enum = schema.get("enum", [])

    # XXX enum can be a dict lookup, handle better
    if schema_enum and value not in schema_enum:
        sinfo = config_info(schema=schema, value=value, source=source)
        raise ConfigInvalidValue(f"{sinfo}\nIs not one of {schema_enum}!")

    if schema_fmt == "password":
        parsed, value = parse_unchanged(value=value)
        if parsed:
            return value

    if isinstance(value, int):
        value = str(value)

    if not isinstance(value, str):
        sinfo = config_info(schema=schema, value=value, source=source)
        raise ConfigInvalidValue(f"{sinfo}\nIs not a string!")

    return value


def config_build(schemas, old_config, new_config, source, check=True, callbacks=None):
    """Pass."""
    for name, schema in schemas.items():
        if name not in new_config and name in old_config:
            new_config[name] = old_config[name]
        elif name in new_config:
            value = new_config[name]
            if check:
                value = config_check(
                    value=value, schema=schema, source=source, callbacks=callbacks
                )
            new_config[name] = value
    return new_config


def config_unknown(schemas, new_config, source, callbacks=None):
    """Pass."""
    unknowns = {k: v for k, v in new_config.items() if k not in schemas}
    if unknowns:
        unknowns = ["{}: {!r}".format(k, v) for k, v in unknowns.items()]
        unknowns = "\n  " + "\n  ".join(unknowns)
        err = f"Unknown settings supplied for {source}: {unknowns}"
        raise ConfigUnknown(tablize_schemas(schemas=schemas, err=err))
    return new_config


def config_unchanged(schemas, old_config, new_config, source, callbacks=None):
    """Pass."""
    if new_config == old_config or not new_config:
        err = f"No changes supplied for {source}"
        raise ConfigUnchanged(
            tablize_schemas(schemas=schemas, config=old_config, err=err)
        )
    return new_config


def config_default(schemas, new_config, source, sane_defaults=None, callbacks=None):
    """Pass."""
    sane_defaults = sane_defaults or {}

    for name, schema in schemas.items():
        if name in new_config:
            continue

        if "default" in schema:
            new_config[name] = schema["default"]

        if name in sane_defaults:
            new_config[name] = sane_defaults[name]

    return new_config


def config_required(schemas, new_config, source, ignores=None, callbacks=None):
    """Pass."""
    missing = []
    ignores = ignores or []
    for name, schema in schemas.items():
        if schema["required"] and name not in new_config and name not in ignores:
            missing.append(schema)

    if missing:
        err = f"Required configurations not supplied for {source}"
        raise ConfigRequired(tablize_schemas(schemas=missing, err=err))
    return new_config


def config_empty(schemas, new_config, source, callbacks=None):
    """Pass."""
    if not new_config:
        err = f"No configuration supplied for {source}"
        raise ConfigRequired(tablize_schemas(schemas=schemas, err=err))
    return new_config


def config_info(schema, value, source):
    """Pass."""
    value_type = type(value).__name__
    return (
        f"Value {value!r} of type {value_type!r} "
        f"supplied for {source} setting {schema['name']!r}"
    )


def is_uploaded_file(value):
    """Pass."""
    check = value

    if isinstance(check, str):
        check = json_load(obj=value, error=False)

    if isinstance(check, dict):
        uuid = check.get("uuid")
        filename = check.get("filename")
        values = [uuid, filename]
        if all(values) and all([isinstance(x, str) for x in values]):
            return True, check
    return False, value


def parse_schema(raw):
    """Pass."""
    parsed = {}
    if not raw:
        return parsed

    schemas = raw["items"]
    required = raw["required"]

    for schema in schemas:
        schema_name = schema["name"]
        parsed[schema_name] = schema
        schema["required"] = schema_name in required

    return parsed


def parse_schema_enum(schema):
    """Pass."""
    # core settings: password_brute_force_protection: conditional
    # has a list of dict enums, so turn it into a lookup map
    if schema.get("enum") and isinstance(schema["enum"][0], dict):
        schema["enum"] = {x["name"]: x for x in schema["enum"]}


def parse_section(raw, raw_config, parent, settings):
    """Pass."""
    # XXX has no title:
    #   settings_gui::saml_login_settings::configure_authncc
    title = raw.get("title", raw["name"].replace("_", " ").title())
    config = raw_config.get(raw["name"], {})
    section_name = raw["name"]
    schemas = raw["items"]
    required = raw["required"]
    section_defaults = raw.get(section_name, {})

    parsed = {}
    parsed["settings_title"] = settings["settings_title"]
    parsed["name"] = section_name
    parsed["title"] = title
    parsed["schemas"] = {}
    parsed["sub_sections"] = sub_sections = {}
    parsed["parent_name"] = parent.get("name", "")
    parsed["parent_title"] = parent.get("title", "")
    parsed["config"] = config

    for schema in schemas:
        parse_schema_enum(schema=schema)
        schema_name = schema["name"]

        items = schema.get("items", None)

        # sub_sections:
        #   {"items": [{}], "required": [""], "type": "array"}
        if isinstance(items, list):
            sub_sections[schema_name] = parse_section(
                raw=schema, raw_config=config, parent=parsed, settings=settings,
            )
            schema.pop("items")

        # non sub_sections:
        #   no items key in schema
        #   {"items": {"type": ""} "type": "array"}
        schema["required"] = schema_name in required
        parsed["schemas"][schema_name] = schema

        # XXX does not follow schema for defaults:
        #   core settings::https_log_settings
        if schema_name in section_defaults and "default" not in schema:
            schema["default"] = section_defaults[schema_name]

        # XXX has no title:
        #   settings_gui::saml_login_settings::configure_authncc
        schema["title"] = schema.get("title", schema_name.replace("_", " ").title())

    return parsed


def parse_settings(raw, title=""):
    """Pass."""
    # XXX missing pretty_name:
    #   settings_gui
    #   settings_lifecycle
    title = raw["schema"].get("pretty_name", "") or title

    raw_config = raw["config"]
    raw_sections = raw["schema"]["items"]

    parsed = {}
    parsed["settings_title"] = title
    parsed["sections"] = sections = {}
    parsed["config"] = copy.deepcopy(raw_config)

    for raw_section in raw_sections:
        section_name = raw_section["name"]
        sections[section_name] = parse_section(
            raw=raw_section, raw_config=raw_config, parent={}, settings=parsed
        )

    return parsed



# -*- coding: utf-8 -*-
"""Test suite for axonius_api_client.tools."""
import os

from click.testing import CliRunner

from ....cli import cli
from ....tools import pathlib


class TestCmdWriteConfig(object):
    """Pass."""

    def test_prompt(self, request, monkeypatch):
        """Pass."""
        runner = CliRunner(mix_stderr=False)

        url = request.config.getoption("--ax-url")
        key = request.config.getoption("--ax-key")
        secret = request.config.getoption("--ax-secret")

        prompt_input = "\n".join([url, key, secret])

        args = ["tools", "write-config"]

        with runner.isolated_filesystem():
            path = pathlib.Path(os.getcwd())
            envfile = path / ".env"
            result1 = runner.invoke(cli=cli, args=args, input=prompt_input)
            assert envfile.is_file()

        exit_code1 = result1.exit_code
        stdout1 = result1.stdout
        stderr1 = result1.stderr

        assert stdout1
        assert stderr1
        assert exit_code1 == 0
